/*	Copyright (C) 1990 Free Software Foundation, Inc.

This file is part of Oleo, the GNU Spreadsheet.

Oleo is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

Oleo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Oleo; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <signal.h>
#ifdef I_IOCTL
#include <sys/ioctl.h>
#endif

#include "funcdef.h"
#define obstack_chunk_alloc ck_malloc
#define obstack_chunk_free free
#include "obstack.h"
#include "sysdef.h"
#include "global.h"
#include "cell.h"
#include "cmd.h"
#include "init.h"
#include "utils.h"
#include "regions.h"
#include "ref.h"
#include "line.h"
#define DEFINE_IO_VARS 1
#include "io_abstract.h"
#include "io_generic.h"
#include "io_utils.h"
#include "io_edit.h"
#include "window.h"
#include "sylk.h"
#include "font.h"

/* External spreadsheet functions */
#ifdef USE_DLD
/* If we're using dynamic linking, we get the names of the
	   functions to call by prepending the basename of save_name onto
		   _read_file
		   _write_file
		   _set_options
		   _show_options
	   so, if the file is sylk.o , the functions are named
		   sylk_read_file
		   sylk_write_file
		   sylk_set_options
		   sylk_show_options
		   */
char *io_name;
#else
#include "sylk.h"
#include "oleofile.h"
#include "sc.h"
#include "list.h"
#endif


const char oleo_version_string[] = "Oleo version 1.1";

#ifdef __STDC__
extern char *getenv ();
extern unsigned ualarm (unsigned int, unsigned int);
#else
extern char *getenv ();
extern unsigned ualarm ();
#endif



#ifdef __STDC__
/* Routines for manipulating key bindings */
static void do_bind_key (struct keymap *, int, int, int);
static void desc_map (struct keymap *);

/* File I/O functions */
static FILE *open_file (char *, struct line *, char *);
static void close_file (struct line *, FILE *);

/* Routines for putting info into 'struct rng's */
static int get_two_ranges (char *, struct rng *, struct rng *, struct line *);
static FILE *get_range_and_file (char *, struct rng *, struct line *);
static char *get_range_and_font (char *, struct rng *, struct line *);
static int get_a_range (char *, struct rng *, struct line *);
static int get_abs_rng (char **, struct rng *);

/* Routines for converting text<-->cell_format */
static int str_to_fmt (char *);

/* Routines for converting text<-->cell_jst */
static int chr_to_jst (int);

/* Routines for implementing user commands */
int global_cmd (int);
void set_options (char *);
void play_cell ();

/* Backends for other functions */
static int do_set_option (char *);


/* Spreadsheet (global) functions declared in this file */


int main (int, char **);
void map_chr (int);
extern void clear_spreadsheet ();
char *cell_name (CELLREF, CELLREF);
char *range_name (struct rng *);
char *fmt_to_str (int);
char *jst_to_str (int);
char *col_to_str (CELLREF);
static void got_sigint (int);
/* The user commands that are defined in this file. . . */
extern void copy_region ();
extern void copy_values_region ();
extern void set_format ();
extern void move_region ();
extern void delete_region ();
extern void format_area ();
extern void open_window ();
extern void close_window ();
extern void hsplit_window ();
extern void vsplit_window ();
extern void delete_window ();
extern void delete_other_windows ();
extern void goto_window ();
extern void other_window ();
extern void print_region (struct rng *, char *);
extern void psprint_region_cmd (FILE *, struct rng *);

static void do_debug (char *);
static void end_macro ();
static void do_break_cmd ();
static void set_default ();
static void quit_cmd ();
static void set_usr_fmt (char *);
static void set_var (char *);
static void show_var (char *);
static void show_all_var ();
static void recalc_cmd ();
static void bind_key_cmd (char *);
static void desc_key_cmd ();
static void read_cmds_cmd (FILE *);
static void write_keys_cmd (FILE *);
static void write_cmd (FILE *);
static void read_cmd (FILE *);
static void read_merge_cmd (FILE *);
static void kill_all_cmd ();
static void sort_region_cmd (char *);
static void write_reg_cmd (FILE *, struct rng *);
static void start_macro ();
void execute_cmd (char *);
static void interact_macro_cmd ();
static void goto_region (struct rng *);
static void mark_cell_cmd ();
static void do_input_cmd (int, int);
static void format_cell_cmd ();
static void kill_cell_cmd ();
static void digit_cmd (int);
static void show_options ();
static void bound_macro (int);
static void make_keymap (char *);
static void set_cell_font_cmd (char *);
static void set_area_font_cmd (char *, struct rng *);
static void scan_cell_cursor (int);
static void shift_cell_cursor (int);
static void scroll_cell_cursor (int);
static void recenter_cur_win ();
static void repaint ();
static void mouse_goto_cmd ();
static void mouse_mark_cmd ();
static void mouse_mark_and_goto_cmd ();
static void set_page_size_cmd (char *);

#ifdef HAVE_X11_X_H
extern void set_x_default_font (char *);
extern void set_x_default_point_size (char *);
#endif
#else
/* Routines for manipulating key bindings */
static void do_bind_key ();
static void desc_map ();

/* File I/O functions */
static FILE *open_file ();
static void close_file ();

/* Routines for putting info into 'struct rng's */
static int get_two_ranges ();
static FILE *get_range_and_file ();
static char *get_range_and_font ();
static int get_a_range ();
static int get_abs_rng ();

/* Routines for converting text<-->cell_format */
static int str_to_fmt ();

/* Routines for converting text<-->cell_jst */
static int chr_to_jst ();

/* Routines for implementing user commands */
int global_cmd ();
void set_options ();
void play_cell ();

/* Backends for other functions */
static int do_set_option ();


/* Spreadsheet (global) functions declared in this file */


int main ();
void map_chr ();
extern void clear_spreadsheet ();
char *cell_name ();
char *range_name ();
char *fmt_to_str ();
char *jst_to_str ();
char *col_to_str ();
static void got_sigint ();
/* The user commands that are defined in this file. . . */
extern void copy_region ();
extern void copy_values_region ();
extern void set_format ();
extern void move_region ();
extern void delete_region ();
extern void format_area ();
extern void open_window ();
extern void hsplit_window ();
extern void vsplit_window ();
extern void delete_window ();
extern void delete_other_windows ();
extern void close_window ();
extern void goto_window ();
extern void other_window ();
extern void print_region ();
extern void psprint_region_cmd ();

static void do_debug ();
static void end_macro ();
static void do_break_cmd ();
static void set_default ();
static void quit_cmd ();
static void set_usr_fmt ();
static void set_var ();
static void show_var ();
static void show_all_var ();
static void recalc_cmd ();
static void bind_key_cmd ();
static void desc_key_cmd ();
static void read_cmds_cmd ();
static void write_keys_cmd ();
static void write_cmd ();
static void read_cmd ();
static void read_merge_cmd ();
static void kill_all_cmd ();
static void sort_region_cmd ();
static void write_reg_cmd ();
static void start_macro ();
void execute_cmd ();
static void interact_macro_cmd ();
static void goto_region ();
static void mark_cell_cmd ();
static void do_input_cmd ();
static void format_cell_cmd ();
static void kill_cell_cmd ();
static void digit_cmd ();
static void show_options ();
static void bound_macro ();
static void make_keymap ();
static void set_cell_font_cmd ();
static void set_area_font_cmd ();
static void scan_cell_cursor ();
static void shift_cell_cursor ();
static void scroll_cell_cursor ();
static void recenter_cur_win ();
static void repaint ();
static void mouse_goto_cmd ();
static void mouse_mark_cmd ();
static void mouse_mark_and_goto_cmd ();
static void set_page_size_cmd ();

#ifdef HAVE_X11_X_H
extern void set_x_default_font ();
extern void set_x_default_point_size ();
#endif

#endif


char *ename[] =
{
  "#WHAT?",
  "#ERROR", "#BAD_INPUT", "#NON_NUMBER", "#NON_STRING",
  "#NON_BOOL", "#NON_RANGE", "#OUT_OF_RANGE", "#NO_VALUES",
  "#DIV_BY_ZERO", "#BAD_NAME", "#NOT_AVAIL", "#PARSE_ERROR",
  "#NEED_OPEN", "#NEED_CLOSE", "#NEED_QUOTE", "#UNK_CHAR",
  "#UNK_FUNC",
  0
};

char tname[] = "#TRUE";
char fname[] = "#FALSE";

struct line val_line;
struct line fmt_line;
struct line wid_line;
struct line hgt_line;

/* Lines 'a' through 'z' */
struct line in_line[26];

unsigned short default_width = 8;
unsigned short default_height = 1;
unsigned short saved_default_width = 8;
unsigned short saved_default_height = 1;
int default_jst = JST_LFT;
int default_fmt = FMT_GEN;
int default_lock = LCK_UNL;

char buf[BUFSIZ];
int term_flag;

extern char print_buf[];

struct keymap **the_maps;
char **map_names;
int num_maps;

struct cmd_func **the_funcs;
int num_funcs;


/* For lines that take text input, the second letter (eg 'ta' or 'fdr')
   specifies which struct line to use to contain the text generated for that
   command.  That way, the command will have the appropriate defaults, etc.
   Currently 'a' through 'n' are used (?).
	a	set options
	b	set/show variables
	c	read/write keymap
	d	read/write file
	e	read-merge	write-region
	f	copy region/copy values region
	g	format-region
	h	move-region
	i	print-region
	j	sort-region
	k	delete-region
	l	execute-command
	m	goto-cell
	n	open-window
	o	close-window
	p	debug
	q	bind-key
	r	set-user-format
	s	create-keymap
 */
static struct cmd_func *end_macro_cmd;
static struct cmd_func *digit_0_cmd;
static struct cmd_func *digit_9_cmd;



static struct cmd_func cmd_funcs[] =
{
  {"unbound", 0, ALL, 0},
/* The location of these motion commands in this array is assumed elsewhere in */
/* the code.  Other motion and scrolling commands are also assumed to be in a */
/* particular order. */
  {"up-cell", "n0", ALL, shift_cell_cursor},
  {"down-cell", "n1", ALL, shift_cell_cursor},
  {"right-cell", "n2", ALL, shift_cell_cursor},
  {"left-cell", "n3", ALL, shift_cell_cursor},
  {"upright-cell", "n4", ALL, shift_cell_cursor},
  {"upleft-cell", "n5", ALL, shift_cell_cursor},
  {"downright-cell", "n6", ALL, shift_cell_cursor},
  {"downleft-cell", "n7", ALL, shift_cell_cursor},

  {"scroll-up", "n0", ALL, scroll_cell_cursor},
  {"scroll-down", "n1", ALL, scroll_cell_cursor},
  {"scroll-right", "n2", ALL, scroll_cell_cursor},
  {"scroll-left", "n3", ALL, scroll_cell_cursor},
  {"scroll-upright", "n4", ALL, scroll_cell_cursor},
  {"scroll-upleft", "n5", ALL, scroll_cell_cursor},
  {"scroll-downright", "n6", ALL, scroll_cell_cursor},
  {"scroll-downleft", "n7", ALL, scroll_cell_cursor},

  {"scan-up", "n0", ALL, scan_cell_cursor},
  {"scan-down", "n1", ALL, scan_cell_cursor},
  {"scan-right", "n2", ALL, scan_cell_cursor},
  {"scan-left", "n3", ALL, scan_cell_cursor},

  {"break", 0, ALL | BRK, do_break_cmd},
  {"recenter-window", 0, ALL, recenter_cur_win},
  {"set-option", "ta", ALL, set_options},
  {"set-defaults", "T", ALL, set_default},
  {"quit", "C", ALL, quit_cmd},
  {"redraw-screen", 0, ALL, repaint},
  {"set-user-format", "trT", ALL, set_usr_fmt},

/* If you change the line used by {set,show}-variable, you must change
   {set,show}_var to use the correct line, else bad things will happen */
  {"set-variable", "tb", ALL, set_var},
  {"show-variable", "tb", ALL, show_var},
  {"show-all-variables", 0, ALL, show_all_var},

  {"recalculate", 0, ALL, recalc_cmd},
  {"bind-key", "tqT", ALL, bind_key_cmd},
  {"describe-key", "T", ALL, desc_key_cmd},
  {"read-commands", "fcr", ALL, read_cmds_cmd},
  {"write-keys", "fcw", ALL, write_keys_cmd},

/* If you change the line used by {read,write}-file, you must change
   main() to use the correct line, else chaos will erupt */
  {"save-spreadsheet", "fdw", ALL, write_cmd},
  {"visit-spreadsheet", "Cfdr", ALL, read_cmd},
  {"merge-spreadsheet", "fer", ALL, read_merge_cmd},
  {"clear-spreadsheet", "C", ALL, kill_all_cmd},

  {"copy-region", "Rf", ALL, copy_region},
  {"copy-values-in_region", "Rf", ALL, copy_values_region},
  {"format-region", "rg", ALL, format_area},
  {"move-region", "Rh", ALL, move_region},
  {"print-region", "Fi", ALL, print_region},
  {"psprint-region", "Fi", ALL, psprint_region_cmd},
  {"set-page-size", "tp", ALL, set_page_size_cmd },
  {"set-default-ps-font", "tq", ALL, set_ps_font_cmd },
#ifdef HAVE_X11_X_H
  {"set-default-font", "to", ALL, set_x_default_font },
  {"set-default-point-size", "tn", ALL, set_x_default_point_size },
#endif

  {"sort-region", "tj", ALL, sort_region_cmd},
  {"write-region-to-file", "Fe", ALL, write_reg_cmd},
  {"delete-region", "rk", ALL, delete_region},

  {"start-entering-macro", 0, ALL, start_macro},
  {"execute-command", "tl", ALL, execute_cmd},
  {"macro-interactive-here", 0, ALL, interact_macro_cmd},

/* If you change which line goto-cell uses, you *must* modify goto_region() */
  {"goto-cell", "rm", ALL, goto_region},
  {"mark-cell", 0, ALL, mark_cell_cmd},
  {"edit-cell", "T0", ALL | WTH_CHR, do_input_cmd},
  {"edit-value-cell", "T1", ALL | WTH_CHR, do_input_cmd},
  {"format-cell", "T", ALL, format_cell_cmd},
  {"delete-cell", 0, ALL, kill_cell_cmd},
  {"edit-cell-with-default", "T3", ALL | WTH_CHR, do_input_cmd},

  {"digit-0", "N", ALL | NC, digit_cmd},
  {"digit-1", "N", ALL | NC, digit_cmd},
  {"digit-2", "N", ALL | NC, digit_cmd},
  {"digit-3", "N", ALL | NC, digit_cmd},
  {"digit-4", "N", ALL | NC, digit_cmd},
  {"digit-5", "N", ALL | NC, digit_cmd},
  {"digit-6", "N", ALL | NC, digit_cmd},
  {"digit-7", "N", ALL | NC, digit_cmd},
  {"digit-8", "N", ALL | NC, digit_cmd},
  {"digit-9", "N", ALL | NC, digit_cmd},

  {"enter-text-in-cell", "T2", WTH_CHR | ALL, do_input_cmd},

  {"show-options", 0, ALL, show_options},
  {"play_cell", 0, ALL, play_cell},
  {"open-window", "tn", ALL, open_window},
  {"split-window-horizontally", 0, ALL, hsplit_window},
  {"split-window-vertically", 0, ALL, vsplit_window},
  {"delete-window", 0, ALL, delete_window},
  {"delete-other-windows", 0, ALL, delete_other_windows},
  {"close-window", "to", ALL, close_window},
  {"goto-window", "to", ALL, goto_window},
  {"other-window", 0, ALL, other_window},

  {"stop-entering-macro", 0, ALL, end_macro},

  {"create-keymap", "ts", ALL, make_keymap},

  {"set-cell-font", "tf", ALL, set_cell_font_cmd},
  {"define-font", "tf", ALL, define_font},
  {"set-region-font", "Sf", ALL, set_area_font_cmd},

  {	"mouse-goto",		0,	ALL,	mouse_goto_cmd },
  {	"mouse-mark",		0,	ALL,	mouse_mark_cmd },
  {	"mouse-mark-and-goto",	0,	ALL,	mouse_mark_and_goto_cmd },
  {"bound-menu", 0, WTH_CHR | ALL, 0},

  {0, 0, 0, 0}

};




#if !defined(SIGIO)
int intr1 = 0;
int intr2 = 0;
int 
real_get_chr ()
{
  static int saved;
  int ret;

  if (saved)
    {
      ret = saved - 1;
      saved = 0;
      goto fini;
    }
  if (auto_recalc)
    {
      current_cycle++;
      if (!bkgrnd_recalc)
	{
	  while (eval_next_cell ())
	    ;
	}
      else
	{
	  int ev;

	  do
	    {
	      io_nodelay (1);
	      while (ev = eval_next_cell ())
		if ((ret = io_getch ()) != -1)
		  goto fini;
#ifdef TEST
		else
		  intr1++;
#endif
	      io_nodelay (0);
	      io_redisp ();
	      ret = io_getch ();
#ifdef TEST
	      if (ret == ERR)
		intr2++;
#endif
	    }
	  while (ret == ERR);
	  goto fini;
	}
    }
  io_update_status ();
  io_redisp ();
  ret = io_getch ();
fini:
  if (making_macro)
    {
      if (ret == 0)
	*making_macro++ = 0x80 | 'a';
      else if (ret == '{')
	*making_macro++ = 0x80 | 'b';
      else
	*making_macro++ = ret;
      if (making_macro >= making_macro_start + making_macro_size)
	{
	  making_macro_start = ck_realloc (making_macro_start, 5 + making_macro_size * 2);
	  making_macro = making_macro_start + making_macro_size;
	  making_macro_size *= 2;
	}
    }
  if (ret & 0x80)
    {
      saved = 1 + (ret & 0x7f);
      return CTRL ('[');
    }
  return ret;
}

#else

#ifndef VOLATILE
#ifdef __STDC__
#define VOLATILE volatile
#else
#define VOLATILE
#endif
#endif
VOLATILE static char ibuf[100];
static int i_in, i_cnt;

int 
real_get_chr ()
{
  int ch;
  static int save;

  if (save)
    {
      ch = save - 1;
      save = 0;
      goto fini;
    }
  if (i_cnt)
    {
      ch = ibuf[i_cnt++];
      if (i_cnt == i_in)
	i_cnt = i_in = 0;
      goto fini;
    }
  for (;;)
    {
      if (auto_recalc)
	{
	  current_cycle++;
	  if (bkgrnd_recalc)
	    {
	      while (!io_input_avail () && eval_next_cell ())
		io_scan_for_input (0);
	    }
	  else
	    {
	      while (eval_next_cell ())
		;
	    }
	}
      io_redisp ();
      if (io_input_avail ())
	{
	  int ret;


	  ret = io_read_kbd (ibuf, sizeof (ibuf));
	  if (ret == 1)
	    {
	      ch = ibuf[0];
	      goto fini;
	    }
	  if (ret > 1)
	    {
	      i_cnt = 1;
	      i_in = ret;
	      ch = ibuf[0];
	      goto fini;
	    }
	  if (ret == 0 || (errno != EWOULDBLOCK && errno != EINTR))
	    return EOF;
	}
      io_wait_for_input ();
    }
fini:
  if (making_macro)
    {
      if (ch == 0)
	*making_macro++ = 0x80 | 'a';
      else if (ch == '{')
	*making_macro++ = 0x80 | 'b';
      else
	*making_macro++ = ch;
      if (making_macro >= making_macro_start + making_macro_size)
	{
	  making_macro_start = ck_realloc (making_macro_start, 5 + making_macro_size * 2);
	  making_macro = making_macro_start + making_macro_size;
	  making_macro_size *= 2;
	}
    }
  if (ch & 0x80)
    {
      save = 1 + (ch & 0x7f);
      ch = CTRL ('[');
    }
  return ch;
}

extern int input_avail_val;

void
got_sigio (sig)
     int sig;
{
  input_avail_val++;
}


#endif




CELLREF setrow, setcol;

CELLREF curow = MIN_ROW, cucol = MIN_COL;

CELLREF mkrow = NON_ROW, mkcol = NON_COL;
int cur_chr;
struct cmd_func *cur_cmd;
short cur_vector;
unsigned int how_many = 1;
char *macro_func_arg = 0;

int n_bound_macros;
struct rng *bound_macros;
int bound_macro_vec;

/* This variable is non-zero if the spreadsheet has been changed in any way */
int modified = 0;

/* User settable options */
int bkgrnd_recalc = 1;
int auto_recalc = 1;
int a0 = 0;
unsigned signal_ticks = 10 * TIMER_MULT;	/* = 10 seconds/tick */

#ifdef __STDC__
void (*read_file) (FILE *, int) = oleo_read_file;
void (*write_file) (FILE *, struct rng *) = oleo_write_file;
int (*set_file_opts) (int, char *) = oleo_set_options;
void (*show_file_opts) () = oleo_show_options;
#else
void (*read_file) () = oleo_read_file;
void (*write_file) () = oleo_write_file;
int (*set_file_opts) () = oleo_set_options;
void (*show_file_opts) () = oleo_show_options;
#endif

const int colmagic[] =
{0, 0, 1, -1, 1, -1, 1, -1, 0};
const int rowmagic[] =
{-1, 1, 0, 0, -1, -1, 1, 1, 0};


struct macro *rmac = 0;
struct obstack macro_stack;
unsigned char *making_macro;
unsigned char *making_macro_start;
unsigned int making_macro_size;


/* Read a character.  If we're in a macro, read from the macro. . . */
int
get_chr ()
{
  int ch;

  if (rmac)
    {
      ch = *(rmac->mac_exe++);
      switch (ch)
	{
	case '{':		/* What else can we do? */
	case '\0':
	  --(rmac->mac_exe);
	  break;

	case (0x80 | 'a'):
	  ch = 0;
	  break;

	case (0x80 | 'b'):
	  ch = '{';
	  break;
	default:
	  break;
	}
    }
  else
    ch = real_get_chr ();
  return ch;
}

void 
open_window (text)
     char *text;
{
  int hv;
  int where;

  while (*text == ' ')
    text++;

  if (*text == 'h' || *text == 'H')
    hv = 0;
  else if (*text == 'v' || *text == 'V')
    hv = 1;
  else
    {
      io_error_msg ("Open 'h'orizontal or 'v'ertical window, not '%s'", text);
      return;
    }
  where = atoi (text + 1);
  while (isspace (*text))
    ++text;
  while (isalnum (*text))
    ++text;
  while (isspace (*text))
    ++text;
  if (*text == '%')
    {
      where *= (hv
		? (cwin->numr + (cwin->lh_wid ? label_rows : 0))
		: (cwin->numc + cwin->lh_wid));
      where /= 100;
    }
  io_win_open (hv, where);
}

void
hsplit_window ()
{
  open_window ("h50%");
}


void
vsplit_window ()
{
  open_window ("v50%");
}


void 
close_window (text)
     char *text;
{
  int num;

  num = atoi (text) - 1;

  if (num < 0 || num >= nwin)
    {
      io_error_msg ("Window %num?", text);
      return;
    }
  if (nwin == 1)
    {
      io_error_msg ("You can't close the last window!");
      return;
    }
  io_win_close (&wins[num]);
}

void 
delete_window ()
{
  io_win_close (cwin);
}

void
delete_other_windows ()
{
  if (nwin > 1)
    {
      CELLREF r = curow;
      CELLREF c = cucol;
      while (nwin > 1)
	io_win_close (cwin);
      io_move_cell_cursor (r, c);
    }
}

void
goto_window (text)
     char *text;
{
  int n;
  n = atoi (text) - 1;
  if (n < 0 || n > nwin)
    {
      io_error_msg ("Window %s doesn't exist.", text);
      return;
    }
  io_set_cwin (&wins[n]);
}

void 
nicely_goto_window (n)
     int n;
{
  if (input_active)
    {
      io_cellize_cursor ();
      window_after_input = n;
      input_active = 0;
      return;
    }
  else
    {
      if (window_after_input == n)
	{
	  io_inputize_cursor ();
	  window_after_input = -1;
	  input_active = 1;
	}
      else
	io_set_cwin (&wins[n]);
    }
}

void 
other_window ()
{
  int n = cwin - wins;
  if (!input_active)
    n = (n + 1) % nwin;
  nicely_goto_window (n);
}

int
set_window_option (set_opt, text)
     int set_opt;
     char *text;
{
  int n;
  int stat;
  static struct opt
    {
      char *text;
      int bits;
    }
  opts[] =
  {
    {
      "reverse", WIN_EDGE_REV
    }
    ,
    {
      "standout", WIN_EDGE_REV
    }
    ,
    {
      "page", WIN_PAG_HZ | WIN_PAG_VT
    }
    ,
    {
      "pageh", WIN_PAG_HZ
    }
    ,
    {
      "pagev", WIN_PAG_VT
    }
    ,
    {
      "lockh", WIN_LCK_HZ
    }
    ,
    {
      "lockv", WIN_LCK_VT
    }
    ,
    {
      "edges", WIN_EDGES
    }
  };
  if ((stat = (!strincmp (text, "status", 6) && isspace (text[6])))
      || (!strincmp (text, "input", 5) && isspace (text[5])))
    {
      int n = set_opt ? atoi (text + 6 + stat) : 0;	/* A little pun. */
      int new_inp = stat ? user_input : n;
      int new_stat = stat ? n : user_status;
      io_set_input_status (new_inp, new_stat, 1);
    }
  else if (!strincmp (text, "link", 4))
    {
      if (set_opt)
	{
	  n = atoi (text + 4) - 1;
	  if (n < 0 || n > nwin)
	    io_error_msg ("Can't '%s': window # out of range", text);
	  else
	    cwin->link = n;
	}
      else
	cwin->link = -1;
    }
  else if (set_opt && !stricmp (text, "unlink"))
    cwin->link = -1;
  else if (set_opt && !strincmp (text, "row ", 4))
    {
      text += 4;
      curow = astol (&text);
    }
  else if (set_opt && !strincmp (text, "col ", 4))
    {
      text += 4;
      cucol = astol (&text);
    }
  else
    {
      for (n = 0; n < sizeof (opts) / sizeof (struct opt); n++)
	if (!stricmp (text, opts[n].text))
	  {
	    if (set_opt)
	      cwin->flags |= opts[n].bits;
	    else
	      cwin->flags &= ~opts[n].bits;
	    break;
	  }

      if (n == sizeof (opts) / sizeof (struct opt))
	  return 0;
    }
  return 1;
}

void
show_window_options ()
{
  int n;

  cwin->curow = curow;
  cwin->cucol = cucol;
  if (user_status)
    io_text_line ("Status line at %d", user_status);
  else
    io_text_line ("Status line disabled.");
  io_text_line ("");
  for (n = 0; n < nwin; n++)
    {
      int flags = wins[n].flags;
      io_text_line ("Window #%d showing %s, with cursor at %s",
		    n + 1,
		    range_name (&wins[n].screen),
		    cell_name (wins[n].curow, wins[n].cucol));
      io_text_line ("   Options:  %sedges (%sreverse)%s%s%s%s",
		    flags & WIN_EDGES ? "" : "no",
		    flags & WIN_EDGE_REV ? "" : "no",
		    flags & WIN_PAG_HZ ? ", pageh" : "",
		    flags & WIN_PAG_VT ? ", pagev" : "",
		    flags & WIN_LCK_HZ ? ", lockh" : "",
		    flags & WIN_LCK_VT ? ", lockv" : "");
      if (wins[n].link != -1)
	io_text_line ("Linked to window %d", wins[n].link + 1);
    }
}

static struct cmd_func *
find_func (table, name)
     struct cmd_func *table;
     char *name;
{
  while (table->func_name)
    if (!strcmp (name, table->func_name))
      return table;
    else
      ++table;
  return 0;
}

void
init_maps ()
{
  num_maps = 0;
  the_maps = 0;
  map_names = 0;

  the_funcs = ck_malloc (sizeof (struct cmd_func *) * 2);
  num_funcs = 2;
  the_funcs[0] = &cmd_funcs[0];
  the_funcs[1] = &edit_funcs[0];

  run_init_cmds ();

  end_macro_cmd = find_func (cmd_funcs, "stop-entering-macro");
  digit_0_cmd = find_func (cmd_funcs, "digit-0");
  digit_9_cmd = find_func (cmd_funcs, "digit-9");
}

static int 
add_usr_cmds (new_cmds)
     struct cmd_func *new_cmds;
{
  num_funcs++;
  the_funcs = ck_realloc (the_funcs, num_funcs * sizeof (struct cmd_func *));
  the_funcs[num_funcs - 1] = new_cmds;
  return num_funcs - 1;
}

#ifdef USE_DLD
static int 
add_usr_maps (new_maps)
     struct keymap **new_maps;
{
  int n;

  for (n = 1; new_maps[n]; n++)
    ;
  the_maps = ck_realloc (the_maps, (n + num_maps) * sizeof (struct keymap *));
  bcopy (new_maps, &the_maps[num_maps], n * sizeof (struct keymap *));
  num_maps += n;
  return num_maps - n;
}

#endif

static int last_map;

static char * disclaimer[] = 
{
  " Copyright (C) 1990, 1991, 1992 Free Software Foundation,Inc.\n",
  "There is ABSOLUTELY NO WARRANTY for Oleo; see the file COPYING\n",
  "for details.  Oleo is free software and you are welcome to distribute\n",
  "copies of it under certain conditions; see the file COPYING to see the\n",
  "conditions.\n\n",
  0
};


int 
main (argc, argv)
     int argc;
     char **argv;
{
  int c;
  int num;
  int using_x = 0;
  void got_sig ();

  myname = argv[0];

  __make_backups = 1;
#if !defined(__TURBOC__) && !defined(SYSV)
  setbuffer (stdout, buf, sizeof (buf));
#endif
  if (argc < 2 || strcmp (argv[1], "-q"))
    {
      if (!getenv ("OLEO"))
	{
	  char * home;
	  char * ptr;
	  FILE * fp;
	  home = getenv ("HOME");
	  if (home)
	    {
	      ptr = mk_sprintf ("%s/%s", home, RCFILE);
	      fp = fopen (ptr, "r");
	      free (ptr);
	      if (fp)
		{
		  fclose (fp);
		  goto skip_disclaimer;
		}
	    }
	  if (fp = fopen (RCFILE, "r"))
	    {
	      fclose (fp);
	      goto skip_disclaimer;
	    }

	  {
	    char ** msg;
	    fputs (oleo_version_string, stdout);
	    for (msg = disclaimer; *msg; ++msg)
	      fputs (*msg, stdout);
	    fflush (stdout);
	    sleep (5);
	  }
	skip_disclaimer:
	}
    }
  else
    {
      --argc;
      ++argv;
    }
  init_infinity ();
  init_mem ();
  init_eval ();
  init_refs ();
  init_cells ();
  init_fonts ();
  obstack_init (&macro_stack);

  c = 0;
  if (argc > 2 && !strcmp (argv[1], "-f"))
    {
      --argc;
      ++argv;
      c++;
    }

#ifdef HAVE_X11_X_H
  if (getenv ("DISPLAY"))
    {
      x11_graphics ();
      using_x = 1;
    }
  else
#endif
    tty_graphics ();
  init_maps ();

  io_open_display ();

  if (argc > 2)
    {
      fprintf (stderr, "Usage: %s [filename]\n", argv[0]);
      exit (1);
    }
#ifdef USE_DLD
  if (!index (myname, '/'))
    {
      char *name;

      name = dld_find_executable (myname);
      num = dld_init (name);
      free (name);
    }
  else
    num = dld_init (myname);
  if (num)
    io_error_msg ("dld_init() failed: %s", (dld_errno < 0 || dld_errno > dld_nerr) ? "Unknown error" : dld_errlst[dld_errno]);
  dld_search_path = ":/usr/local/lib/oleo:/lib:/usr/lib:/usr/local/lib";
#endif


  {
    FILE *fp;
    extern int fclose ();
    char *ptr, *home;
    /* would be nice to use ck_fopen, but these files might not
       exist. . . */

    home = getenv ("HOME");
    if (home)
      {
	ptr = mk_sprintf ("%s/%s", home, RCFILE);
	if (fp = fopen (ptr, "r"))
	  {
	    read_cmds_cmd (fp);
	    fclose (fp);
	  }
	free (ptr);
      }

    if (fp = fopen (RCFILE, "r"))
      {
	read_cmds_cmd (fp);
	fclose (fp);
      }
    if (argc == 2)
      {
	set_line (&in_line[3], argv[1]);
	if (fp = fopen (argv[1], "r"))
	  {
	    (*read_file) (fp, 0);
	    fclose (fp);
	  }
	else
	  io_error_msg ("Can't open %s: %s", argv[1], err_msg ());
      }
  }

  signal (SIGINT, got_sigint);
#ifdef SIGQUIT
  signal (SIGQUIT, got_sig);
  signal (SIGILL, got_sig);
  signal (SIGEMT, got_sig);
  signal (SIGBUS, got_sig);
  signal (SIGSEGV, got_sig);
  signal (SIGPIPE, got_sig);
#endif
  if (!using_x)
    {
#if  defined(SIGIO)
      signal (SIGIO, got_sigio);
#ifdef FASYNC
      term_flag = fcntl (0, F_GETFL, FASYNC /* |FNDELAY */ );
      c = fcntl (0, F_SETFL, FASYNC /* |FNDELAY */ );
      if (c == -1)
	{
	  fprintf (stderr, "Can not turn on fasync or fndelay on stdin: %s\n",
		   err_msg ());
	  exit (1);
	}
#else
#ifdef FIOSSAIOSTAT
      {
	int i;
	
	ioctl (0, FIOGSAIOSTAT, &i);
	term_flag = i;
	ioctl (0, FIOGNBIO, &i);
	if (i)
	  term_flag += 2;
	i = 1;
	ioctl (0, FIOSSAIOSTAT, &i);
	/* ioctl(0,FIOSNBIO,&i); */
	i = getpid ();
	ioctl (0, FIOSSAIOOWN, &i);
      }
#endif
#endif
#endif
    } /* if (!using_x) */
  io_recenter_cur_win ();

  for (;;)
    {
      if (how_many != 1)
	{
	  how_many = 1;
	  io_update_status ();
	}
      io_clear_input_before ();
      map_chr (MAIN_MAP);
      if (!rmac)
	io_clear_input_after ();

    swtch:
      if (!cur_cmd)
	{
	  io_error_msg ("Key %s is unbound in map %s",
			char_to_string (cur_chr),
			map_names[last_map]);
	  continue;
	}
      if ((cur_cmd->func_flags & NONTOP) == 0)
	{
	  io_error_msg ("Command '%s' is not appropriate now.", cur_cmd->func_name);
	  continue;
	}
      num = global_cmd (MAIN_MAP);
      if (num == -2)
	io_error_msg ("Command '%s' is *not* appropriate now.", cur_cmd->func_name);
      else if (num >= 0)
	{
	  c = num;
	  goto swtch;
	}
    }
}

static FILE *
open_file (prompt, line, mode)
     char *prompt;
     struct line *line;
     char *mode;
{
  FILE *fp;

  if (io_get_line (prompt, line))
    return 0;
  if ((fp = xopen_with_backup (line->buf, mode)) == 0)
    {
      io_error_msg ("Can't open file '%s':%s", line->buf, err_msg ());
      return 0;
    }
  return fp;
}

static void
close_file (line, fp)
     struct line *line;
     FILE *fp;
{
  int num;

  if (num = xclose (fp))
    io_error_msg ("Can't close '%s': Error code %d: %s", line->buf, num, err_msg ());
}

/* Deal with point-n-shoot */
static int
get_a_range (whatfor, r, line)
     char *whatfor;
     struct rng *r;
     struct line *line;
{
  char *ptr;

  if (mkrow != NON_ROW)
    {
      /* The range has already been selected */
    got_rng:
      set_rng (r, mkrow, mkcol, curow, cucol);
      mkrow = NON_ROW;
      return 0;
    }
  switch (io_get_line (whatfor, line))
    {
    case 0:
      if (mkrow != NON_ROW)
	goto got_rng;
      ptr = line->buf;
      if (get_abs_rng (&ptr, r) || *ptr != '\0')
	{
	  io_error_msg ("Can't parse '%s'", line->buf);
	  return 1;
	}
      return 0;
    case 1:
      if (mkrow != NON_ROW)
	goto got_rng;
      return 1;
    case 2:
      return 1;
#ifdef TEST
    default:
      panic ("io_get_line() what?");
#endif
    }
  return 1;
}

static char *
get_range_and_font (cmd_name, r, line)
     char *cmd_name;
     struct rng *r;
     struct line *line;
{
  char *ptr;
  char *prm;
  int got_rng = 0;
  extern char *strdup ();

  if (mkrow != NON_ROW)
    {
      /* We've got the range, just get the string */
      set_rng (r, mkrow, mkcol, curow, cucol);
      got_rng++;
      mkrow = NON_ROW;
      prm = mk_sprintf ("%s %s to font", cmd_name, range_name (r));
    }
  else
    prm = strdup (cmd_name);

  if (io_get_line (prm, line))
    {
      free (prm);
      return 0;
    }
  free (prm);

  if (mkrow != NON_ROW)
    {
      if (got_rng)
	{
	  io_error_msg ("Two ranges?");
	  return 0;
	}
      set_rng (r, mkrow, mkcol, curow, cucol);
      mkrow = NON_ROW;
      got_rng++;
    }
  ptr = line->buf;
  if (!got_rng)
    {
      if (get_abs_rng (&ptr, r))
	{
	  io_error_msg ("Can't parse '%s'", line->buf);
	  return 0;
	}
      if (*ptr == ',')
	ptr++;
    }
  while (*ptr == ' ')
    ptr++;
  if (!*ptr)
    {
      if (io_get_line (cmd_name, line))
	{
	  return 0;
	}
      ptr = line->buf;
    }
  return ptr;
}


static FILE *
get_range_and_file (whatfor, r, line)
     char *whatfor;
     struct rng *r;
     struct line *line;
{
  char *ptr;
  char *prompt;
  int got_rng = 0;
  FILE *ret;
  extern char *strdup ();

  if (mkrow != NON_ROW)
    {
      /* We've got the range, just get the file */
      set_rng (r, mkrow, mkcol, curow, cucol);
      got_rng++;
      mkrow = NON_ROW;
      prompt = mk_sprintf ("%s %s to file", whatfor, range_name (r));
    }
  else
    prompt = strdup (whatfor);

  if (io_get_line (prompt, line))
    {
      free (prompt);
      return 0;
    }
  free (prompt);

  if (mkrow != NON_ROW)
    {
      if (got_rng)
	{
	  io_error_msg ("Two ranges?");
	  return 0;
	}
      set_rng (r, mkrow, mkcol, curow, cucol);
      mkrow = NON_ROW;
      got_rng++;
    }
  ptr = line->buf;
  if (!got_rng)
    {
      if (get_abs_rng (&ptr, r))
	{
	  io_error_msg ("Can't parse '%s'", line->buf);
	  return 0;
	}
      if (*ptr == ',')
	ptr++;
    }
  while (*ptr == ' ')
    ptr++;
  if (!*ptr)
    {
      if (io_get_line ("To file: ", line))
	return 0;
      ptr = line->buf;
    }
  ret = xopen_with_backup (ptr, "w");
  if (!ret)
    io_error_msg ("Can't open file '%s':%s", ptr, err_msg ());
  return ret;
}


static int
get_two_ranges (whatfor, one, two, line)
     char *whatfor;
     struct rng *one;
     struct rng *two;
     struct line *line;
{
  char *ptr;
  char *prompt;
  int num_got = 0;
  int n;

  prompt = 0;
  if (mkrow != NON_ROW)
    {
      /* We've got the first one already */
      set_rng (one, mkrow, mkcol, curow, cucol);
      mkrow = NON_ROW;
      num_got++;
    }

  while (num_got != 2)
    {
      prompt = (num_got == 1) ? mk_sprintf ("%s %s to", whatfor, range_name (one)) : whatfor;
      n = io_get_line (prompt, line);
      if (prompt != whatfor)
	free (prompt);

      if (n > 1 || (n == 1 && mkrow == NON_ROW))
	return 1;

      if (mkrow != NON_ROW)
	{
	  if (line->buf[0])
	    {
	      io_error_msg (num_got ? "Extra characters '%s'" : "'%s' is ambiguous", line->buf);
	      return 1;
	    }
	  set_rng (num_got ? two : one, mkrow, mkcol, curow, cucol);
	  mkrow = NON_ROW;

	  num_got++;
	}
      else
	{
	  ptr = line->buf;
	  if (!num_got)
	    {
	      if (get_abs_rng (&ptr, one))
		{
		  io_error_msg ("Can't parse first range in '%s'", line->buf);
		  return 1;
		}
	      if (*ptr == ',')
		ptr++;
	    }

	  if (get_abs_rng (&ptr, two))
	    {
	      if (num_got)
		{
		  io_error_msg ("Can't find second range in '%s'", line->buf);
		  return 1;
		}
	      num_got++;
	    }
	  else if (!num_got)
	    num_got += 2;
	  else
	    num_got++;
	}
    }
  if (*ptr)
    {
      io_error_msg ("Extra characters in '%s'", line->buf);
      return 1;
    }

  sprint_line (line, "%s %s", range_name (one), range_name (two));
  return 0;
}


void 
map_chr (map)
     int map;
{
  int ch;
  struct keymap *keymap;
  struct key *key;

  keymap = the_maps[map];
  last_map = map;
  for (;;)
    {
      if (rmac)
	{
	  int len;
	  unsigned char *ptr;

	tryagain:
	  ch = *(rmac->mac_exe++);
	  switch (ch)
	    {
	    case '\0':
	      cur_vector = 0;
	      cur_cmd = end_macro_cmd;
	      cur_chr = 0;
	      return;

	    case 0x80 | 'a':
	      ch = '\0';
	      break;

	    case 0x80 | 'b':
	      ch = '{';
	      break;

	    case '{':
	      for (ptr = rmac->mac_exe;
		   *ptr && *ptr != ' ' && *ptr != '}';
		   ptr++);
	      len = ptr - rmac->mac_exe;
	      for (cur_vector = 0; cur_vector < num_funcs; cur_vector++)
		for (cur_cmd = &the_funcs[cur_vector][0];
		     cur_cmd->func_name;
		     cur_cmd++)
		  if (!strincmp ((char *) (rmac->mac_exe),
				 cur_cmd->func_name, len)
		      && cur_cmd->func_name[len] == '\0')
		    {
		      cur_chr = '\0';
		      goto out;
		    }
	      io_error_msg ("Ignoring unknown function '%.*s' in macro",
			    len, rmac->mac_exe);
	      while (*ptr != '\0' && *ptr != '}')
		ptr++;
	      if (*ptr == '}')
		ptr++;
	      rmac->mac_exe = ptr;
	      goto tryagain;

	    out:
	      if (*ptr == ' ')
		{
		  /* ... add argument support here ... */
		  if (!cur_cmd->func_args)
		    {
		      io_error_msg ("Ignoring extra operand to %s",
				    cur_cmd->func_name);
		      while (*ptr && *ptr != '}')
			ptr++;
		      if (*ptr == '}')
			ptr++;
		    }
		  else if (cur_cmd->func_args[0] == 'n')
		    {
		      how_many = astol ((char **) (&ptr));
		      if (*ptr == '}')
			ptr++;
		    }
		  else
		    {
		      macro_func_arg = (char *) ptr;
		      while (*ptr && *ptr != '}')
			ptr++;
		      if (*ptr == '}')
			*ptr++ = '\0';
		    }
		  rmac->mac_exe = ptr;
		}
	      else
		rmac->mac_exe += len + 1;
	      return;
	    }
	}
      else
	ch = real_get_chr ();

      for (;;)
	{
	  key = &(keymap->keys[ch]);
	  if (key->vector < 0)
	    if (key->code >= 0)
	      {
		keymap = the_maps[key->code];
		last_map = key->code;
		break;
	      }
	    else if (keymap->map_next)
	      keymap = keymap->map_next;
	    else
	      {
		cur_vector = 0;
		cur_cmd = 0;
		cur_chr = ch;
		return;
	      }
	  else
	    {
	      cur_vector = key->vector;
	      cur_cmd = &(the_funcs[key->vector][key->code]);
	      cur_chr = ch;
	      return;
	    }
	}
    }
}

/* Return values:
   -3	break key!
   -2	c is inappropriate
   -1	C is OK
   0+	New char value is ret-1;
   */

int 
global_cmd (magic)
     int magic;
{
  char *ptr;
  struct rng fm, to;
  FILE *fp;
  static struct line confirm;
  struct cmd_func *cmd;

  if (!cur_cmd)
    return -2;
  cmd = cur_cmd;
  if (cmd->func_args)
    {
      ptr = cmd->func_args;

    next_arg:
      switch (*ptr++)
	{
	case 'R':
	  if (get_two_ranges (cmd->func_name, &fm, &to, &in_line[*ptr++ - 'a']))
	    break;
	  (*cmd->func_func) (&fm, &to);
	  break;

	case 'r':
	  if (get_a_range (cmd->func_name, &to, &in_line[*ptr++ - 'a']))
	    break;
	  (*cmd->func_func) (&to);
	  break;

	case 't':
	  if (io_get_line (cmd->func_name, &in_line[*ptr++ - 'a']))
	    break;
	  (*cmd->func_func) (in_line[cmd->func_args[1] - 'a'].buf);
	  break;

	case 'C':
	  if (modified == 0)
	    goto next_arg;
	  set_line (&confirm, "");
	  if (io_get_line ("Are you SURE", &confirm)
	      || (confirm.buf[0] != 'Y' && confirm.buf[0] != 'y'))
	    return -1;
	  goto next_arg;

	case 'f':
	  fp = open_file (cmd->func_name, &in_line[*ptr - 'a'], ptr + 1);
	  if (!fp)
	    break;
	  (*cmd->func_func) (fp);
	  close_file (&in_line[*ptr++ - 'a'], fp);
	  break;

	case 'F':
	  fp = get_range_and_file (cmd->func_name, &to, &in_line[*ptr - 'a']);
	  if (!fp)
	    break;
	  (*cmd->func_func) (fp, &to);
	  close_file (&in_line[*ptr++ - 'a'], fp);
	  break;

	case 'S':
	  {
	    char *cp = get_range_and_font (cmd->func_name, &to, &in_line[*ptr - 'a']);
	    if (!cp)
	      break;
	    (*cmd->func_func) (cp, &to);
	    break;
	  }

	case 'n':
	  if (!*ptr)
	    (*cmd->func_func) ();
	  else if (*ptr >= '0' && *ptr <= '9')
	    (*cmd->func_func) (*ptr - '0');
	  else if (*ptr >= 'a' && *ptr <= 'z')
	    (*cmd->func_func) ((*ptr - 'a') * 26 + ptr[1] - 'a');
	  else
	    (*cmd->func_func) ();
	  break;

	case 'T':
	  if (*ptr >= '0' && *ptr <= '9')
	    (*cmd->func_func) (*ptr - '0', cur_chr);
	  else
	    (*cmd->func_func) ();
	  break;

	case 'N':
	  (*cmd->func_func) (magic);
	  return cur_chr;

	case '\0':
	  (*cmd->func_func) ();
	  break;

#ifdef TEST
	default:
	  panic ("Unknown arg type %s in global_cmd", ptr);
#endif
	}
      if (cmd->func_flags & BRK)
	return -3;
      return -1;
    }
  else if (cmd->func_func)
    {
      if (cmd->func_flags & WTH_CHR)
	(*cmd->func_func) (cur_chr);
      else
	(*cmd->func_func) ();
      if (cmd->func_flags & BRK)
	return -3;
      return -1;
    }
  return -2;
}

static void 
write_keys_cmd (fp)
     FILE *fp;
{
  struct keymap *map;
  int n;
  int key;
  int vec;
  int code;

  for (n = 0; n < num_maps; n++)
    {
      char *def;
      map = the_maps[n];
      def = 0;
      if (map && map->map_next)
	{
	  for (key = 0; key < num_maps; key++)
	    if (the_maps[key] == map->map_next)
	      {
		def = map_names[key];
		break;
	      }
	}
      if (def)
	fprintf (fp, "create-keymap %s %s\n", map_names[n], def);
      else
	fprintf (fp, "create-keymap %s\n", map_names[n]);
    }
  for (n = 0; n < num_maps; n++)
    {
      map = the_maps[n];
      for (key = 0; key < 256; key++)
	{
	  vec = map->keys[key].vector;
	  code = map->keys[key].code;
	  if (vec < 0 && code >= 0)
	    fprintf (fp, "bind-key %s %s %s\n",
		     map_names[n],
		     map_names[code],
		     char_to_string (key));
	  else if (vec >= 0)
	    fprintf (fp, "bind-key %s %s %s\n",
		     map_names[n],
		     the_funcs[vec][code].func_name,
		     char_to_string (key));
	}
    }
}

static void 
clear_keymap (m)
     struct keymap *m;
{
  int n;
  for (n = 0; n < 256; n++)
    {
      m->keys[n].vector = -1;
      m->keys[n].code = -1;
    }
}

int 
map_id (name)
     char *name;
{
  int x;
  for (x = 0; x < num_maps; ++x)
    if (!strcmp (name, map_names[x]))
      return x;
  return -1;
}

static void 
make_keymap (mapname)
     char *mapname;
{
  char *ptr;
  int num;
  struct keymap *linkmap;
  extern char *strdup ();

  for (ptr = mapname; *ptr && !isspace (*ptr); ++ptr);
  linkmap = 0;
  if (*ptr)
    {
      *ptr++ = '\0';
      for (num = 0; num < num_maps; num++)
	{
	  if (!strcmp (map_names[num], ptr))
	    {
	      linkmap = the_maps[num];
	      break;
	    }
	}
    }
  for (num = 0; num < num_maps; num++)
    {
      if (!strcmp (map_names[num], mapname))
	{
	  clear_keymap (the_maps[num]);
	  return;
	}
    }
  the_maps = ck_realloc (the_maps, (num_maps + 1) * sizeof (struct keymap *));
  the_maps[num_maps] = ck_malloc (sizeof (struct keymap));
  the_maps[num_maps]->map_next = linkmap;
  {
    int c;
    struct key k;
    k.code = k.vector = -1;
    for (c = 0; c < 256; ++c)
      the_maps[num_maps]->keys[c] = k;
  }
  map_names = ck_realloc (map_names, (num_maps + 1) * sizeof (char *));
  map_names[num_maps] = strdup (mapname);
  num_maps++;
}

static void 
read_cmds_cmd (fp)
     FILE *fp;
{
  char *ptr;
  while (fgets (print_buf, 300, fp))
    {
      if (ptr = (char *)rindex (print_buf, '\n'))
	*ptr = '\0';
      for (ptr = print_buf; isspace (*ptr); ptr++);
      if (!*ptr || (*ptr == '#'))
	continue;
      execute_cmd (ptr);
    }
}

static void
write_cmd (fp)
     FILE *fp;
{
  (*write_file) (fp, 0);
  modified = 0;
}

static void
read_cmd (fp)
     FILE *fp;
{
  (*read_file) (fp, 0);
}

static void
read_merge_cmd (fp)
     FILE *fp;
{
  (*read_file) (fp, 1);
}

static void
write_reg_cmd (fp, rng)
     FILE *fp;
     struct rng *rng;
{
  (*write_file) (fp, rng);
}

static void
sort_region_cmd (ptr)
     char *ptr;
{
  struct cmp
    {
      char mult;
      CELLREF row;
      CELLREF col;
    };

  struct rng tmp_rng;

  extern struct rng sort_rng;
  extern struct rng sort_ele;
  extern struct cmp *sort_keys;
  extern int sort_keys_num;
  extern int sort_keys_alloc;
  extern void sort_region ();

  if (get_abs_rng (&ptr, &sort_rng))
    {
      io_error_msg ("Can't find a range to sort in %s", ptr);
      return;
    }

  cur_row = sort_rng.lr;
  cur_col = sort_rng.lc;

  while (*ptr == ' ')
    ptr++;
  if (!*ptr)
    {
      sort_ele.lr = 0;
      sort_ele.lc = 0;
      sort_ele.hr = 0;
      sort_ele.hc = 0;
    }
  else if (!parse_cell_or_range (&ptr, &sort_ele))
    {
      io_error_msg ("Can't parse elements in %s", ptr);
      return;
    }
  else
    {
      sort_ele.lr -= sort_rng.lr;
      sort_ele.lc -= sort_rng.lc;
      sort_ele.hr -= sort_rng.lr;
      sort_ele.hc -= sort_rng.lc;
    }

  sort_keys_num = 0;
  while (*ptr == ' ')
    ptr++;
  for (; *ptr;)
    {
      if (sort_keys_num == sort_keys_alloc)
	{
	  sort_keys_alloc++;
	  if (sort_keys_alloc > 1)
	    sort_keys = ck_realloc (sort_keys, sort_keys_alloc * sizeof (struct cmp));
	  else
	    sort_keys = ck_malloc (sizeof (struct cmp));
	}
      sort_keys[sort_keys_num].mult = 1;
      if (*ptr == '+')
	ptr++;
      else if (*ptr == '-')
	{
	  sort_keys[sort_keys_num].mult = -1;
	  ptr++;
	}
      if (!*ptr)
	{
	  sort_keys[sort_keys_num].row = 0;
	  sort_keys[sort_keys_num].col = 0;
	  sort_keys_num++;
	  break;
	}
      if (!parse_cell_or_range (&ptr, &tmp_rng) || tmp_rng.lr != tmp_rng.hr || tmp_rng.lc != tmp_rng.hc)
	{
	  io_error_msg ("Can't parse key #%d in %s", sort_keys_num + 1, ptr);
	  sort_keys_num = -1;
	  return;
	}
      sort_keys[sort_keys_num].row = tmp_rng.lr - sort_rng.lr;
      sort_keys[sort_keys_num].col = tmp_rng.lc - sort_rng.lc;
      sort_keys_num++;

      while (*ptr == ' ')
	ptr++;
    }
  if (sort_keys_num == 0)
    {
      if (sort_keys_alloc == 0)
	{
	  sort_keys_alloc++;
	  sort_keys = ck_malloc (sizeof (struct cmp));
	}
      sort_keys[0].mult = 1;
      sort_keys[0].row = 0;
      sort_keys[0].col = 0;
      sort_keys_num++;
    }
  sort_region ();
  io_repaint ();
}

static void
set_var (ptr)
     char *ptr;
{
  int num;
  char *ret;

  while (*ptr == ' ')
    ptr++;
  for (num = 0; ptr[num] && ptr[num] != ' '; num++)
    ;
  modified = 1;
  if (!ptr[num])
    ret = new_var_value (ptr, num, (char *) 0);
  else
    ret = new_var_value (ptr, num, &ptr[num + 1]);
  if (ret)
    io_error_msg ("Can't set-variable %s: %s\n", ptr, ret);
}

static void
show_var (ptr)
     char *ptr;
{
  struct var *v;
  int num;

  while (*ptr == ' ')
    ptr++;
  for (num = 0; ptr[num] && ptr[num] != ' '; num++)
    ;

  v = find_var (ptr, num);
  if (!v || v->var_flags == VAR_UNDEF)
    {
      io_error_msg ("There is no '%s'", ptr);
      return;
    }
  if (a0)
    {
      if (v->v_rng.lr != v->v_rng.hr || v->v_rng.lc != v->v_rng.hc)
	/* FOO */ sprintf (print_buf, "%s $%s$%u:$%s$%u", v->var_name, col_to_str (v->v_rng.lc), v->v_rng.lr, col_to_str (v->v_rng.hc), v->v_rng.hr);
      else
	/* FOO */ sprintf (print_buf, "%s $%s$%u", v->var_name, col_to_str (v->v_rng.lc), v->v_rng.lr);
    }
  else
    sprintf (print_buf, "%s %s", v->var_name, range_name (&(v->v_rng)));
  io_info_msg (print_buf);
  set_line (&in_line[1], print_buf);
}

static void
show_a_var (name, v)
     char *name;
     struct var *v;
{
  if (v->var_flags == VAR_UNDEF)
    return;
  if (a0)
    {
      if (v->v_rng.lr != v->v_rng.hr || v->v_rng.lc != v->v_rng.hc)
	/* FOO */ io_text_line ("%-20s  $%s$%u:$%s$%u", v->var_name, col_to_str (v->v_rng.lc), v->v_rng.lr, col_to_str (v->v_rng.hc), v->v_rng.hr);
      else
	/* FOO */ io_text_line ("%-20s  $%s$%u", v->var_name, col_to_str (v->v_rng.lc), v->v_rng.lr);
    }
  else
    io_text_line ("%-20s  %s", v->var_name, range_name (&(v->v_rng)));
}

static void
show_all_var ()
{
  io_text_start ();
  io_text_line ("%-20s  Current Value", "Variable Name");
  for_all_vars (show_a_var);
  io_text_finish ();
}

static void
shift_cell_cursor (dir)
     int dir;
{
  io_shift_cell_cursor (dir);
}


static void
scroll_cell_cursor (dir)
     int dir;
{
  io_scroll_cell_cursor (dir);
}

static void
repaint ()
{
  io_repaint ();
}

static void
recenter_cur_win ()
{
  io_recenter_cur_win ();
}


static void
scan_cell_cursor (magic)
     int magic;
{
  CELLREF pos, pos_end, pos_inc;
  CELL *cp;
  int over, down;


  over = colmagic[magic] * how_many;
  down = rowmagic[magic] * how_many;
  if (over)
    {
      if (over > 0)
	{
	  pos = max_col (curow);
	  pos_end = MIN_COL;
	  pos_inc = -1;
	}
      else
	{
	  pos = MIN_COL;
	  pos_end = max_col (curow);
	  pos_inc = 1;
	}
      for (;;)
	{
	  for (; ((!(cp = find_cell (curow, pos)) || !GET_TYP (cp))); pos += pos_inc)
	    {
	      if (pos == pos_end)
		{
		  how_many = 1;
		  pos = MIN_COL;
		  break;
		}
	    }
	  if (--how_many == 0)
	    break;
	  pos += pos_inc;
	}
      io_move_cell_cursor (curow, pos);
    }
  else
    {
      if (down > 0)
	{
	  pos = max_row (cucol);
	  pos_end = MIN_ROW;
	  pos_inc = -1;
	}
      else
	{
	  pos = MIN_ROW;
	  pos_end = max_row (cucol);
	  pos_inc = 1;
	}
      for (;;)
	{
	  for (; (!(cp = find_cell (pos, cucol)) || !GET_TYP (cp)); pos += pos_inc)
	    {
	      if (pos == pos_end)
		{
		  how_many = 1;
		  pos = MIN_ROW;
		  break;
		}
	    }
	  if (--how_many == 0)
	    break;
	  pos += pos_inc;
	}

      io_move_cell_cursor (pos, cucol);
    }
}


char *
fmt_to_str (fmt)
     int fmt;
{
  char *ptr;
  static char buf[30];
  char nbuf[10];

  if (fmt == FMT_DEF)
    return "default";
  if (fmt == FMT_HID)
    return "hidden";
  if (fmt == FMT_GPH)
    return "graph";
  if ((fmt & PRC_FLT) == PRC_FLT)
    strcpy (nbuf, "float");
  else
    sprintf (nbuf, "%d", (fmt & PRC_FLT));
  switch (fmt | PRC_FLT)
    {
    case FMT_USR:
      ptr = "user-";
      sprintf (nbuf, "%d", (fmt & PRC_FLT) + 1);
      break;
    case FMT_GEN:
      ptr = "general.";
      break;
    case FMT_DOL:
      ptr = "dollar.";
      break;
    case FMT_CMA:
      ptr = "comma.";
      break;
    case FMT_PCT:
      ptr = "percent.";
      break;
    case FMT_FXT:
      if ((fmt & PRC_FLT) == 0)
	return "integer";
      if (fmt == FMT_FXT)
	return "decimal";
      ptr = "fixed.";
      break;
    case FMT_EXP:
      ptr = "exponent.";
      break;
    default:
      io_error_msg ("Unknown format %d (%x)", fmt, fmt);
      ptr = "UNKNOWN";
      break;
    }
  sprintf (buf, "%s%s", ptr, nbuf);
  return buf;
}

struct fmt
{
  int fmt;
  char **strs;
};

static char *def_names[] =
{"default", "def", "D", 0};
static char *hid_names[] =
{"hidden", "hid", "H", 0};
static char *gph_names[] =
{"graph", "gph", "*", 0};
static char *int_names[] =
{"integer", "int", "I", 0};
static char *dec_names[] =
{"decimal", "dec", 0};

static struct fmt simple[] =
{
  {FMT_DEF, def_names},
  {FMT_HID, hid_names},
  {FMT_GPH, gph_names},
  {FMT_FXT - PRC_FLT, int_names},
  {FMT_FXT, dec_names},
  {0, 0}
};

char *gen_names[] =
{"general.", "gen.", "G", 0};
char *dol_names[] =
{"dollar.", "dol.", "$", 0};
char *cma_names[] =
{"comma.", "com.", ",", 0};
char *pct_names[] =
{"percent.", "pct.", "%", 0};
char *fxt_names[] =
{"fixed.", "fxt.", "F", 0};
char *exp_names[] =
{"exponent.", "exp.", "E", 0};

static struct fmt withprec[] =
{
  {FMT_GEN - PRC_FLT, gen_names},
  {FMT_DOL - PRC_FLT, dol_names},
  {FMT_CMA - PRC_FLT, cma_names},
  {FMT_PCT - PRC_FLT, pct_names},
  {FMT_FXT - PRC_FLT, fxt_names},
  {FMT_EXP - PRC_FLT, exp_names},
  {0, 0}
};

static int
str_to_fmt (ptr)
     char *ptr;
{
  struct fmt *f;
  char **strs;
  int n;
  int ret;
  char *p1, *p2;

  for (f = simple; f->strs; f++)
    {
      for (strs = f->strs; *strs; strs++)
	{
	  if (*ptr != **strs)
	    continue;
	  for (p1 = ptr, p2 = *strs; *p1 == *p2 && *p1; p1++, p2++)
	    ;
	  if (*p1 == '\0' && *p2 == '\0')
	    return f->fmt;
	}
    }
  if (!strncmp (ptr, "user-", 5))
    {
      ptr += 5;
      n = astol (&ptr);
      if (*ptr || n < 1 || n > 16)
	return -1;
      return n - 1 - PRC_FLT + FMT_USR;
    }
  for (f = withprec, ret = 0; !ret && f->strs; f++)
    {
      for (strs = f->strs; *strs; strs++)
	{
	  if (*ptr != **strs)
	    continue;
	  for (p1 = ptr, p2 = *strs; *p2 && *p1 == *p2; p1++, p2++)
	    ;
	  if (!*p2)
	    {
	      ret = f->fmt;
	      ptr = p1;
	      break;
	    }
	}
    }

  if (!ret || !*ptr)
    return -1;
  if (!strcmp (ptr, "float") || !strcmp (ptr, "f"))
    {
      n = PRC_FLT;
    }
  else
    {
      n = astol (&ptr);
      if (*ptr || n < 0 || n > 14)
	return -1;
    }
  return ret + n;
}

char *
jst_to_str (jst)
     int jst;
{
  if (jst == JST_DEF)
    return "default";
  if (jst == JST_LFT)
    return "left";
  if (jst == JST_RGT)
    return "right";
  if (jst == JST_CNT)
    return "center";
  return "unknown";
}

static int
chr_to_jst (chr)
     int chr;
{
  if (chr == 'd' || chr == 'D')
    return JST_DEF;
  if (chr == 'l' || chr == 'L')
    return JST_LFT;
  if (chr == 'r' || chr == 'R')
    return JST_RGT;
  if (chr == 'c' || chr == 'C')
    return JST_CNT;
  return -1;
}

/* parse a range, then turn it into an absolute rng */
static int
get_abs_rng (pptr, retp)
     char **pptr;
     struct rng *retp;
{
  unsigned char n;

  while (**pptr == ' ')
    (*pptr)++;
  if (!**pptr)
    return 1;
  cur_row = curow;
  cur_col = cucol;
  n = parse_cell_or_range (pptr, retp);
  if (!n)
    {
      struct var *v;
      char *ptr;

      ptr = *pptr;
      while (ptr[n] && ptr[n] != ' ')
	n++;
      v = find_var (ptr, n);
      if (!v)
	return 1;
      (*pptr) += n;
      *retp = v->v_rng;
    }
  return 0;
}

static void 
desc_map (map)
     struct keymap *map;
{
  int n;

  io_text_start ();
  for (n = 0; n < 256; n++)
    if (map->keys[n].code >= 0)
      io_text_line ("%s: %d %d %s",
		    char_to_string (n),
		    map->keys[n].vector,
		    map->keys[n].code,
		    (map->keys[n].vector < 0
		     ? "vector"
		     : (the_funcs[map->keys[n].vector]
			[map->keys[n].code].func_name)));
  io_text_finish ();
}

static void 
set_usr_fmt (fmtstr)
     char *fmtstr;
{
  int u_num;
  char *ptr;
  static struct line usr_line;
  struct line tmp_line;
  char *tmp_ptr;
  char *data_buf[9];
  int i;
  static char *names[9] =
  {
    "Positive header",
    "Negative header",
    "Positive trailer",
    "Negative trailer",
    "Zero",
    "Comma",
    "Decimal point",
    "Precision",
    "Scale-factor"
  };

  tmp_line.buf = ck_malloc (10);
  tmp_line.alloc = 10;
  tmp_ptr = tmp_line.buf;

  ptr = fmtstr;
  u_num = astol (&ptr);
  if (u_num < 1 || u_num > 16 || *ptr != '\0')
    {
      io_error_msg ("Unknown number %s", fmtstr);
      return;
    }
  --u_num;

  get_usr_stats (u_num, data_buf);
  for (i = 0; i < 9; i++)
    {
      int slen;
      int prevlen;

      set_line (&usr_line, data_buf[i]);
      if (io_get_line (names[i], &usr_line) > 1)
	return;

      slen = strlen (usr_line.buf);
      if (tmp_line.alloc <= (tmp_ptr - tmp_line.buf) + slen)
	{
	  prevlen = tmp_ptr - tmp_line.buf;
	  tmp_line.alloc += slen + 1;
	  tmp_line.buf = ck_realloc (tmp_line.buf, tmp_line.alloc);
	  tmp_ptr = tmp_line.buf + prevlen;
	}
      strcpy (tmp_ptr, usr_line.buf);
      data_buf[i] = tmp_ptr;
      while (*tmp_ptr != '\0')
	tmp_ptr++;
      tmp_ptr++;
    }

  set_usr_stats (u_num, data_buf);
  free (tmp_line.buf);
  io_repaint ();
}


void
read_mp_usr_fmt (ptr)
     char *ptr;
{
  int usr_n = -1;
  int n_chrs = 0;
  char *p;
  char *buf[9];
  int i;

  for (i = 0; i < 9; i++)
    buf[i] = "";
  p = ptr;
  while (*p == ';')
    {
      *p++ = '\0';
      switch (*p++)
	{
	case 'N':
	  usr_n = astol (&p) - 1;
	  break;
	case 'H':
	  switch (*p++)
	    {
	    case 'P':
	      i = 0;
	      break;
	    case 'N':
	      i = 1;
	      break;
	    default:
	      goto badline;
	    }
	  goto count_chars;
	case 'T':
	  switch (*p++)
	    {
	    case 'P':
	      i = 2;
	      break;
	    case 'N':
	      i = 3;
	      break;
	    default:
	      goto badline;
	    }
	  goto count_chars;

	case 'Z':
	  i = 4;
	  goto count_chars;

	case 'C':
	  i = 5;
	  goto count_chars;

	case 'D':
	  i = 6;
	  goto count_chars;

	case 'P':
	  i = 7;
	  goto count_chars;

	case 'S':
	  i = 8;
	  goto count_chars;

	count_chars:
	  buf[i] = p;
	  n_chrs++;
	  while (*p && *p != ';')
	    {
	      p++;
	      n_chrs++;
	    }
	  break;

	default:
	badline:
	  io_error_msg ("Unknown OLEO line %s", ptr);
	  return;
	}
    }
  if (*p || usr_n < 0 || usr_n > 15)
    goto badline;

  set_usr_stats (usr_n, buf);
}

/* Modify this to write out *all* the options */
void
write_mp_options (fp)
     FILE *fp;
{
  fprintf (fp, "O;%sauto;%sbackground;%sa0;ticks %d\n",
	   auto_recalc ? "" : "no",
	   bkgrnd_recalc ? "" : "no",
	   a0 ? "" : "no",
	   signal_ticks / TIMER_MULT);
}

void 
read_mp_options (str)
     char *str;
{
  char *np;

  while (np = (char *)index (str, ';'))
    {
      *np = '\0';
      do_set_option (str);
      *np++ = ';';
      str = np;
    }
  if (np = (char *)rindex (str, '\n'))
    *np = '\0';
  (void) do_set_option (str);
}

void
set_options (ptr)
     char *ptr;
{
  if (do_set_option (ptr))
    io_recenter_cur_win ();
}

void
play_cell ()
{
  CELL *cp;
  static char buf[20];
  int ch;

  cp = find_cell (curow, cucol);
  if (cp
      && ((GET_LCK (cp) == LCK_DEF && default_lock == LCK_LCK)
	  || GET_LCK (cp) == LCK_LCK))
    {
      io_error_msg ("Cell %s is locked", cell_name (curow, cucol));
      return;
    }
  io_info_msg ("Now playing %s.", cell_name (curow, cucol));
  while (1)
    {
      ch = real_get_chr ();
      if (ch == EOF || ch == '\027')
	return;
      sprintf (buf, "%d", ch);
      new_value (curow, cucol, buf);
      modified = 1;
    }
}

static void 
show_options ()
{
  int n;
  int fmts;
  char *data_buf[9];

  extern char *strdup ();
  extern void show_window_options ();
  extern int usr_set_fmts ();

  n = auto_recalc;
  io_text_start ();

  io_text_line ("auto-recalculation: %s        Recalculate in background: %s",
		n ? " on" : "off", bkgrnd_recalc ? "on" : "off");
  io_text_line ("make backup files:  %s        Copy files into backups:   %s",
	__make_backups ? " on" : "off", __backup_by_copying ? "on" : "off");

  io_text_line ("Asynchronous updates every %u ???",
		signal_ticks / TIMER_MULT);

  io_text_line ("Print width:      %5u", print_width);

  io_text_line ("");

  (*show_file_opts) ();

  io_text_line ("");
  show_window_options ();
  io_text_line ("");

  fmts = usr_set_fmts ();
  if (fmts)
    {
      io_text_line ("User-defined formats:");
      io_text_line ("Fmt    +Hdr    -Hdr   +Trlr   -Trlr    Zero   Comma Decimal  Prec         Scale");
      for (n = 0; n < 16; n++)
	{
	  if (fmts & (1 << n))
	    {
	      get_usr_stats (n, data_buf);
	      io_text_line ("%3d %7s %7s %7s %7s %7s %7s %7s %5s %13s",
			    n + 1,
			    data_buf[0],
			    data_buf[1],
			    data_buf[2],
			    data_buf[3],
			    data_buf[4],
			    data_buf[5],
			    data_buf[6],
			    data_buf[7],
			    data_buf[8]);
	    }
	}
    }
  else
    io_text_line ("No user-defined formats have been defined");

  io_text_finish ();
}

static int 
do_set_option (ptr)
     char *ptr;
{
  int set_opt = 1;
  extern int set_window_option ();

  while (*ptr == ' ')
    ptr++;
  if (!strincmp ("no", ptr, 2))
    {
      ptr += 2;
      set_opt = 0;
      while (*ptr == ' ')
	ptr++;
    }
  if (!stricmp ("auto", ptr))
    {
      auto_recalc = set_opt;
      return 0;
    }
  if (!stricmp ("bkgrnd", ptr) || !stricmp ("background", ptr))
    {
      bkgrnd_recalc = set_opt;
      return 0;
    }
  if (!stricmp ("a0", ptr))
    {
      a0 = set_opt;
      io_repaint ();
      return 0;
    }
  if (!stricmp ("backup", ptr))
    {
      __make_backups = set_opt;
      return 0;
    }
  if (!stricmp ("bkup_copy", ptr))
    {
      __backup_by_copying = set_opt;
      return 0;
    }
  if (set_opt && !strincmp ("ticks ", ptr, 6))
    {
      ptr += 6;
      signal_ticks = astol (&ptr) * TIMER_MULT;
#ifndef __TURBOC__
      if (timer_active)
	ualarm (signal_ticks, signal_ticks);
#endif
      return 0;
    }
  if (set_opt && !strincmp ("print ", ptr, 6))
    {
      ptr += 6;
      print_width = astol (&ptr);
      return 0;
    }
  if (set_opt && !strincmp ("file ", ptr, 5))
    {
#ifdef USE_DLD
      char *tmpstr;
      extern char *strdup ();

      ptr += 5;
      tmpstr = ck_malloc (strlen (ptr) + 20);
      if (io_name)
	{
	  sprintf (tmpstr, "%s.o", ptr);
	  if (dld_unlink_by_file (tmpstr, 0))
	    {
	      io_error_msg ("Couldn't unlink old file format %s: %s", io_name, (dld_errno < 0 || dld_errno > dld_nerr) ? "Unknown error" : dld_errlst[dld_errno]);
	      goto bad_file;
	    }
	  free (io_name);
	}
      if (!stricmp (ptr, "panic"))
	{
	  io_name = 0;
	  read_file = panic_read_file;
	  write_file = panic_write_file;
	  set_file_opts = panic_set_options;
	  show_file_opts = panic_show_options;
	  free (tmpstr);
	  return 0;
	}
      io_name = strdup (ptr);
      sprintf (tmpstr, "%s.o", ptr);
      if (dld_link (tmpstr))
	{
	  io_error_msg ("Couldn't link new file format %s: %s", io_name, (dld_errno < 0 || dld_errno > dld_nerr) ? "Unknown error" : dld_errlst[dld_errno]);
	  goto bad_file;
	}
      if (dld_link ("libc.a"))
	io_error_msg ("Couldn't link libc.a");
      if (dld_link ("libm.a"))
	io_error_msg ("Couldn't link libm.a");

      sprintf (tmpstr, "%s_read_file", ptr);
      read_file = dld_function_executable_p (tmpstr) ? dld_get_func (tmpstr) : 0;
      sprintf (tmpstr, "%s_write_file", ptr);
      write_file = dld_function_executable_p (tmpstr) ? dld_get_func (tmpstr) : 0;

      sprintf (tmpstr, "%s_set_options", ptr);
      set_file_opts = (int (*)()) (dld_function_executable_p (tmpstr) ? dld_get_func (tmpstr) : 0);
      sprintf (tmpstr, "%s_show_options", ptr);
      show_file_opts = dld_function_executable_p (tmpstr) ? dld_get_func (tmpstr) : 0;

      if (!read_file
	  || !write_file
	  || !set_file_opts
	  || !show_file_opts)
	{
	  char **missing;
	  int n;

	  missing = dld_list_undefined_sym ();
	  io_text_start ();
	  io_text_line ("Undefined symbols in file format %s:", ptr);
	  io_text_line ("");
	  for (n = 0; n < dld_undefined_sym_count; n++)
	    io_text_line ("%s", missing[n]);
	  io_text_line ("");
	  io_text_finish ();
	  free (missing);
	  io_error_msg ("File format %s has undefined symbols: not loaded", ptr);
	bad_file:
	  sprintf (tmpstr, "%s.o", io_name);
	  dld_unlink_by_file (io_name, 0);
	  if (io_name)
	    free (io_name);
	  io_name = 0;
	  read_file = panic_read_file;
	  write_file = panic_write_file;
	  set_file_opts = panic_set_options;
	  show_file_opts = panic_show_options;
	}
      free (tmpstr);
#else
      ptr += 5;
      if (!stricmp ("oleo", ptr))
	{
	  read_file = oleo_read_file;
	  write_file = oleo_write_file;
	  set_file_opts = oleo_set_options;
	  show_file_opts = oleo_show_options;
	}
      else if (!stricmp ("sylk", ptr))
	{
	  read_file = sylk_read_file;
	  write_file = sylk_write_file;
	  set_file_opts = sylk_set_options;
	  show_file_opts = sylk_show_options;
	}
      else if (!stricmp ("sc", ptr))
	{
	  read_file = sc_read_file;
	  write_file = sc_write_file;
	  set_file_opts = sc_set_options;
	  show_file_opts = sc_show_options;
	}
      else if (!stricmp ("panic", ptr))
	{
	  read_file = panic_read_file;
	  write_file = panic_write_file;
	  set_file_opts = panic_set_options;
	  show_file_opts = panic_show_options;
	}
      else if (!stricmp ("list", ptr))
	{
	  read_file = list_read_file;
	  write_file = list_write_file;
	  set_file_opts = list_set_options;
	  show_file_opts = list_show_options;
	  /*if (ptr[4])
	    {
	    ptr+=4;
	    sl_sep=string_to_char(&ptr);
	    } */
	}
      else
	io_error_msg ("Unknown file format %s", ptr);
#endif
      return 0;
    }
#ifdef USE_DLD
  else if (!strincmp (ptr, "load ", 5))
    {
      char *tmpstr;
      struct function *new_funs;
      struct cmd_func *new_cmds;
      struct keymap **new_maps;
      void (*init_cmd) ();

      extern unsigned long dld_get_symbol (char *);
      extern void add_usr_funs (struct function *);

      ptr += 5;
      tmpstr = ck_malloc (strlen (ptr) + 20);
      sprintf (tmpstr, "%s.o", ptr);
      if (dld_link (tmpstr))
	{
	  io_error_msg ("Couldn't link %s: %s", tmpstr, (dld_errno < 0 || dld_errno > dld_nerr) ? "Unknown error" : dld_errlst[dld_errno]);
	  free (tmpstr);
	  return 0;
	}
      if (dld_link ("libc.a"))
	io_error_msg ("Couldn't link libc.a");
      if (dld_link ("libm.a"))
	io_error_msg ("Couldn't link libm.a");

      if (dld_undefined_sym_count)
	{
	  char **missing;
	  int n;

	  missing = dld_list_undefined_sym ();
	  io_text_start ();
	  io_text_line ("Undefined symbols in file format %s:", ptr);
	  io_text_line ("");
	  for (n = 0; n < dld_undefined_sym_count; n++)
	    io_text_line ("%s", missing[n]);
	  io_text_line ("");
	  io_text_finish ();
	  free (missing);
	  io_error_msg ("%d undefined symbols in %s", dld_undefined_sym_count, ptr);
	  dld_unlink_by_file (tmpstr, 0);
	  free (tmpstr);
	  return 0;
	}
      sprintf (tmpstr, "%s_funs", ptr);
      new_funs = (struct function *) dld_get_symbol (tmpstr);
      if (new_funs)
	add_usr_funs (new_funs);
      sprintf (tmpstr, "%s_cmds", ptr);
      new_cmds = (struct cmd_func *) dld_get_symbol (tmpstr);
      if (new_cmds)
	add_usr_cmds (new_cmds);
      sprintf (tmpstr, "%s_maps", ptr);
      new_maps = (struct keymap **) dld_get_symbol (tmpstr);
      if (new_maps)
	add_usr_maps (new_maps);
      if (!new_funs && !new_cmds && !new_maps)
	{
	  io_error_msg ("Couldn't find anything to load in %s", ptr);
	  sprintf (tmpstr, "%s.o", ptr);
	  dld_unlink_by_file (tmpstr, 0);
	}
      sprintf (tmpstr, "%s_init", ptr);
      init_cmd = dld_function_executable_p (tmpstr) ? dld_get_func (tmpstr) : 0;
      if (init_cmd)
	(*init_cmd) ();
      free (tmpstr);
      return 0;
    }
#endif
  if (set_window_option (set_opt, ptr) == 0)
    {
      if ((*set_file_opts) (set_opt, ptr))
	io_error_msg ("Unknown option '%s'", ptr);
      return 0;
    }
  return 1;
}


void 
execute_cmd (str)
     char *str;
{
  CELL *cp;
  struct macro *old;
  struct rng r;
  char *ptr = str;

  while (isspace (*str))
    ++str;
  if (!*str || *str == '#')
    return;
  for (ptr = str; *ptr && *ptr != ' '; ptr++);

  if (*ptr)
    *ptr++ = '\0';
  else
    ptr = 0;

  for (cur_vector = 0; cur_vector < num_funcs; cur_vector++)
    for (cur_cmd = &the_funcs[cur_vector][0]; cur_cmd->func_name; cur_cmd++)
      if (!stricmp (str, cur_cmd->func_name))
	{
	  /* these lines stolen from map_char */
	  if (ptr)
	    {
	      if (!cur_cmd->func_args)
		io_error_msg ("Ignoring extra operand to %s",
			      cur_cmd->func_name);
	      else if (cur_cmd->func_args[0] == 'n')
		how_many = astol ((char **) (&ptr));
	      else
		macro_func_arg = ptr;
	    }
	  cur_chr = '\0';
	  global_cmd (MAIN_MAP);
	  return;
	}

  if (get_abs_rng (&str, &r))
    {
      io_error_msg ("Unknown command %s", str);
      return;
    }
  if (ptr)
    {
      io_error_msg ("Macros can't take arguments");
      return;
    }

  if (!(cp = find_cell (r.lr, r.lc))
      || GET_TYP (cp) != TYP_STR
      || cp->cell_str[0] == '\0')
    return;

  old = rmac;
  rmac = (struct macro *) obstack_alloc (&macro_stack, sizeof (struct macro));
  rmac->mac_prev = old;
  rmac->mac_rng = r;
  rmac->mac_row = r.lr;
  rmac->mac_col = r.lc;
  (void) obstack_grow (&macro_stack, cp->cell_str, 1 + strlen (cp->cell_str));
  rmac->mac_exe = (unsigned char *) obstack_finish (&macro_stack);
  rmac->mac_flags = 0;
}

static void
goto_region (r)
     struct rng *r;
{
  struct rng tmp;

  if (mkrow != NON_ROW)
    {
      CELLREF cx, cy;

      set_rng (&tmp, curow, cucol, mkrow, mkcol);
      set_line (&in_line[12], range_name (&tmp));
      cx = mkrow;
      mkrow = curow;
      cy = mkcol;
      mkcol = cucol;
      (void) io_move_cell_cursor (cx, cy);
    }
  else
    {
      set_line (&in_line[12], cell_name (curow, cucol));
      (void) io_move_cell_cursor (r->lr, r->lc);
    }
  if (r->hr != r->lr || r->hc != r->lc)
    {
      mkrow = r->hr;
      mkcol = r->hc;
      io_update_status ();
    }
  else if (mkrow != NON_ROW)
    {
      mkrow = NON_ROW;
      io_update_status ();
    }
}

static void
set_default ()
{
  int fun;
  char *ptr;
  int num;

  sprint_line (&wid_line, "%u", default_width);
  sprint_line (&hgt_line, "%u", default_height);

  set_line (&fmt_line, fmt_to_str (default_fmt));
  io_info_msg ("Alignment %s   Format %s   %slocked  Width %u  Height %u",
	       jst_to_str (default_jst),
	       fmt_line.buf,
	       default_lock == LCK_LCK ? "" : "un",
	       default_width,
	       default_height);

  fun = io_get_chr ("[A]lignment, [F]ormat, [P]rotection [W]idth, or [H]eight  ");
  switch (fun)
    {
    case 'w':
    case 'W':
      if (io_get_line ("set-default-width", &wid_line))
	break;
      ptr = wid_line.buf;
      num = astol (&ptr);
      if (num < 1)
	io_error_msg ("Can't set default width to '%s'", wid_line.buf);
      else
	{
	  default_width = num;
	  io_recenter_all_win ();
	  return;
	}
      break;

    case 'h':
    case 'H':
      if (io_get_line ("set-default height", &hgt_line))
	break;
      ptr = hgt_line.buf;
      num = astol (&ptr);
      if (num < 1)
	io_error_msg ("Can't set default height to '%s'", hgt_line.buf);
      else
	{
	  default_height = num;
	  io_recenter_all_win ();
	  return;
	}
      break;

    case 'p':
    case 'P':
    case 'l':
    case 'L':
      fun = io_get_chr ("[P]rotected, or [U]nprotected");
      if (fun == 'p' || fun == 'P')
	default_lock = LCK_LCK;
      else if (fun == 'u' || fun == 'U')
	default_lock = LCK_UNL;
      else			/* if (main_map[fun]!=BREAK_CMD) */
	io_error_msg ("Unknown char '%s'", char_to_string (fun));
      break;

    case 'f':
    case 'F':
      if (io_get_line ("set-default-format", &fmt_line))
	break;
      num = str_to_fmt (fmt_line.buf);
      if (num == -1 || num == FMT_DEF)
	{
	  io_error_msg ("Unknown format '%s'", fmt_line.buf);
	  break;
	}
      default_fmt = num;
      io_repaint ();
      return;

    case 'a':
    case 'A':
      fun = io_get_chr ("[L]eft, [R]ight, or [C]enter");
      num = chr_to_jst (fun);
      if (num != -1 && num != JST_DEF)
	{
	  default_jst = num;
	  io_repaint ();
	  return;
	}
      else			/* if (main_map[fun]!=BREAK_CMD) */
	io_error_msg ("Unknown Alignment '%s'", char_to_string (fun));
      break;

    default:
      /* if (main_map[fun]!=BREAK_CMD) */
      io_error_msg ("Unknown command '%s'", char_to_string (fun));
      /* else
	 io_error_msg(""); */
      break;
    }
  io_update_status ();
}

void 
format_area (f)
     struct rng *f;
{
  int c;
  int fmt, jst, wid, hgt;
  CELLREF cc;
  char *locked;
  CELL *cp;
  int fun;
  char *ptr;

  cp = find_cell (f->lr, f->lc);
  if (!cp)
    {
      fmt = FMT_DEF;
      jst = JST_DEF;
    }
  else
    {
      fmt = GET_FMT (cp);
      jst = GET_JST (cp);
    }

  wid = get_nodef_width (f->lc);
  if (wid == 0)
    set_line (&wid_line, "def");
  else
    sprint_line (&wid_line, "%d", wid - 1);

  hgt = get_nodef_height (f->lr);
  if (hgt == 0)
    set_line (&hgt_line, "def");
  else
    sprint_line (&hgt_line, "%d", hgt - 1);

  set_line (&fmt_line, fmt_to_str (fmt));
  if (!cp || GET_LCK (cp) == LCK_DEF)
    locked = (default_lock == LCK_UNL ? "unlocked(def)" : "locked(def)");
  else if (GET_LCK (cp) == LCK_UNL)
    locked = "unlocked";
  else if (GET_LCK (cp) == LCK_LCK)
    locked = "locked";
  else
    locked = "Huh What?";

  io_info_msg ("Alignment %s   Format %s   Width %s   Height %s  %s",
	       jst_to_str (jst),
	       fmt_line.buf,
	       wid_line.buf,
	       hgt_line.buf,
	       locked);

  fun =
    io_get_chr
    ("[A]lignment, [F]ormat, [P]rotection, [H]eight or [W]idth  ");
  switch (fun)
    {
    case 'f':
    case 'F':
      if (io_get_line ("set-format", &fmt_line))
	break;
      fmt = str_to_fmt (fmt_line.buf);
      if (fmt != -1)
	format_region (f, fmt, -1);
      else
	io_error_msg ("Unknown format '%s'", fmt_line.buf);
      break;

    case 'a':
    case 'A':
      c = io_get_chr ("Align [L]eft, [R]ight, [C]enter, or [D]efault");
      fun = chr_to_jst (c);
      if (fun != -1)
	format_region (f, -1, fun);
      else			/* if (main_map[c]!=BREAK_CMD) */
	io_error_msg ("Unknown Justify '%s'", char_to_string (c));
      break;

    case 'p':
    case 'P':
    case 'l':
    case 'L':
      c = io_get_chr ("[D]efault, [P]rotect, or [U]nprotect");
      if (c == 'd' || c == 'D')
	lock_region (f, LCK_DEF);
      else if (c == 'p' || c == 'P')
	lock_region (f, LCK_LCK);
      else if (c == 'u' || c == 'U')
	lock_region (f, LCK_UNL);
      else			/* if (main_map[c]!=BREAK_CMD) */
	io_error_msg ("Unknown lock %s", char_to_string (c));
      break;

    case 'w':
    case 'W':
      if (io_get_line ("set-width", &wid_line))
	break;
      ptr = wid_line.buf;
      if (*ptr == 'd' || *ptr == 'D')
	fun = 0;
      else if (isdigit (*ptr))
	fun = astol (&ptr) + 1;
      else
	{
	  io_error_msg ("Unknown width '%s'", wid_line.buf);
	  break;
	}
      for (cc = f->lc;; cc++)
	{
	  set_width (cc, fun);
	  if (cc == f->hc)
	    break;
	}
      io_recenter_all_win ();
      return;
    case 'h':
    case 'H':
      if (io_get_line ("set height", &hgt_line))
	break;
      ptr = hgt_line.buf;
      if (*ptr == 'd' || *ptr == 'D')
	fun = 0;
      else
	{
	  fun = astol (&ptr) + 1;
	  if (*ptr)
	    {
	      io_error_msg ("Unknown height '%s'", hgt_line.buf);
	      break;
	    }
	}
      for (cc = f->lr;; cc++)
	{
	  set_height (cc, fun);
	  if (cc == f->hr)
	    break;
	}
      io_recenter_all_win ();
      return;

    default:
      /* if (main_map[fun]!=BREAK_CMD) */
      io_error_msg ("Unknown command '%s'", char_to_string (fun));
      break;
    }
  io_update_status ();
}


void
got_sig ()
{
  FILE *fp;
  extern int write ();

  /* Using xopen_with_backup is almost certainly a bad idea, however
     using open() (preferred) or fopen() causes trouble in the restore,
     so we suffer. . . */

  fp = xopen_with_backup ("oleo.panic", "w");
  if (fp)
    {
      write (1, "Performing panic save to 'oleo.panic'\r\n", 39);
      panic_write_file (fp, 0);
      xclose (fp);
    }
  else
    write (1, "Can't open 'oleo.panic'!  You lose!\r\n", 37);
  io_close_display ();
#ifdef FASYNC
  fcntl (0, F_SETFL, term_flag);
#endif
#ifdef FIOSSAIOSTAT
  if ((term_flag & 1) == 0)
    {
      int i = 0;

      ioctl (0, FIOSSAIOSTAT, &i);
    }
  if ((term_flag & 2) == 0)
    {
      int i = 0;

      ioctl (0, FIOSNBIO, &i);
    }
#endif

  exit (11);
}

static void
got_sigint (ign)
     int ign;
{
  int ch;

  ch = io_get_chr ("Panic save and exit?");
  if (ch == 'y' || ch == 'Y')
    got_sig ();
}


static void
start_macro ()
{
  if (making_macro)
    {
      io_error_msg ("Can't define two macros at once");
      return;
    }
  making_macro_size = 20;
  making_macro = making_macro_start = ck_malloc (5 + making_macro_size);
  /* *making_macro++='"'; */
}

static void
end_macro ()
{
  union vals z;
  struct rng to;
  CELL *cp;
  struct macro *old;
  static struct line macro_line;

  if (!rmac && !making_macro)
    {
      io_error_msg ("Not executing or defining a macro!");
      return;
    }
  if (rmac)
    {
      if (rmac->mac_row == rmac->mac_rng.hr && rmac->mac_col == rmac->mac_rng.hc)
	{
	  old = rmac->mac_prev;
	  (void) obstack_free (&macro_stack, rmac);
	  rmac = old;
	  goto deal_making;
	}

      if (rmac->mac_row == rmac->mac_rng.hr)
	{
	  rmac->mac_row = rmac->mac_rng.lr;
	  rmac->mac_col++;
	}
      else
	rmac->mac_row++;
      if (!(cp = find_cell (rmac->mac_row, rmac->mac_col)) || GET_TYP (cp) != TYP_STR || cp->cell_str[0] == '\0')
	{
	  old = rmac->mac_prev;
	  (void) obstack_free (&macro_stack, rmac);
	  rmac = old;
	  goto deal_making;
	}
      (void) obstack_grow (&macro_stack, cp->cell_str, 1 + strlen (cp->cell_str));
      rmac->mac_exe = (unsigned char *) obstack_finish (&macro_stack);
      rmac->mac_flags = 0;
    }
deal_making:
  if (!making_macro)
    return;


  making_macro[0] = '\0';
  making_macro = 0;
  if (get_a_range ("Put macro where", &to, &macro_line))
    io_error_msg ("Forgetting new macro");
  else
    {
      z.c_s = (char *) making_macro_start;
      set_new_value (to.lr, to.lc, TYP_STR, &z);
    }
  free (making_macro_start);
}

static void
quit_cmd ()
{
  io_close_display ();
#ifdef FASYNC
  fcntl (0, F_SETFL, term_flag);
#endif
#ifdef FIOSSAIOSTAT
  if ((term_flag & 1) == 0)
    {
      int i = 0;

      ioctl (0, FIOSSAIOSTAT, &i);
    }
  if ((term_flag & 2) == 0)
    {
      int i = 0;

      ioctl (0, FIOSNBIO, &i);
    }
#endif
  exit (0);
}

static void
bound_macro (num)
     int num;
{
  struct macro *old;
  CELL *cp;
  /* CELLREF rr,cc; */

  cp = find_cell (bound_macros[num].lr, bound_macros[num].lc);
  if (!cp || GET_TYP (cp) != TYP_STR || cp->cell_str[0] == '\0')
    return;
  old = rmac;
  rmac = (struct macro *) obstack_alloc (&macro_stack, sizeof (struct macro));
  rmac->mac_prev = old;
  rmac->mac_rng = bound_macros[num];
  rmac->mac_row = bound_macros[num].lr;
  rmac->mac_col = bound_macros[num].lc;
  (void) obstack_grow (&macro_stack, cp->cell_str, 1 + strlen (cp->cell_str));
  rmac->mac_exe = (unsigned char *) obstack_finish (&macro_stack);
  rmac->mac_flags = 0;
}

static void
mark_cell_cmd ()
{
  mkrow = curow;
  mkcol = cucol;
  io_update_status ();
}

static void
recalc_cmd ()
{
  current_cycle++;
  while (eval_next_cell ())
    ;
}

static void
desc_key_cmd ()
{
  io_error_msg ("Key:  ");
  map_chr (MAIN_MAP);
  if (!cur_cmd)
    io_info_msg ("%s is unbound", char_to_string (cur_chr));
  else
    io_info_msg ("%s --> %s", char_to_string (cur_chr), cur_cmd->func_name);
}

static void 
bind_key_cmd (text)
     char *text;
{
  int map;
  int c;
  short vec;
  short code;
  char *ptr;
  char *cmdstr;
  struct rng rng;
  struct cmd_func *tmpfunc;
  static struct line tmpline;

  for (ptr = text; *ptr && !isspace (*ptr); ptr++);
  if (!*ptr)
    {
      if (io_get_line ("Command", &tmpline))
	return;
      ptr = tmpline.buf;
    }
  else
    *ptr++ = '\0';
  while (*ptr == ' ')
    ptr++;
  cmdstr = ptr;
  while (*ptr && *ptr != ' ')
    ptr++;
  if (!*ptr)
    c = io_get_chr ("Key:  ");
  else
    {
      char *save = ptr;
      *ptr++ = '\0';
      c = string_to_char (&ptr);
      if (c < 0)
	{
	  io_error_msg ("Illegal character '%s'", ptr);
	  goto fail;
	}
    }

  for (map = 0; map < num_maps; map++)
    if (!strcmp (text, map_names[map]))
      break;
  if (map == num_maps)
    {
      io_error_msg ("Can't find keymap '%s'", text);
      goto fail;
    }
  for (vec = 0; vec < num_funcs; vec++)
    {
      for (code = 0; the_funcs[vec][code].func_name; code++)
	if (!stricmp (cmdstr, the_funcs[vec][code].func_name))
	  goto fini;
    }
  for (code = 0; code < num_maps; code++)
    {
      if (!strcmp (cmdstr, map_names[code]))
	{
	  vec = -1;
	  goto fini;
	}
    }
  if (get_abs_rng (&cmdstr, &rng))
    {
      io_error_msg ("Unknown command '%s'", cmdstr);
      goto fail;
    }

  vec = bound_macro_vec;
  for (code = 0; code < n_bound_macros; code++)
    {
      if (!bcmp (&bound_macros[code], &rng, sizeof (struct rng)))
	  goto fini;
    }

  ptr = range_name (&rng);
  if (!bound_macro_vec)
    {
      bound_macros = ck_malloc (sizeof (struct rng));
      n_bound_macros = 1;
      tmpfunc = ck_malloc (sizeof (struct cmd_func));
      bound_macro_vec = add_usr_cmds (tmpfunc);
    }
  else
    {
      n_bound_macros++;
      bound_macros = ck_realloc (bound_macros, n_bound_macros * sizeof (struct rng));
      the_funcs[bound_macro_vec] = ck_realloc (the_funcs[bound_macro_vec], n_bound_macros * sizeof (struct cmd_func));
      tmpfunc = &the_funcs[bound_macro_vec][n_bound_macros - 1];
    }
  bound_macros[n_bound_macros - 1] = rng;
  tmpfunc->func_args = ck_malloc (strlen (ptr) + 5);
  tmpfunc->func_args[0] = 'n';
  tmpfunc->func_args[1] = 'a' + (n_bound_macros - 1) / 26;
  tmpfunc->func_args[2] = 'a' + (n_bound_macros - 1) % 26;
  tmpfunc->func_args[3] = '\0';
  tmpfunc->func_name = tmpfunc->func_args + 4;
  strcpy (tmpfunc->func_name, ptr);
  tmpfunc->func_flags = ALL;
  tmpfunc->func_func = bound_macro;
fini:
  do_bind_key (the_maps[map], c, vec, code);
fail:
  sprint_line (&in_line[16], "%s %s %s", text, cmdstr, char_to_string (c));
}

static void 
do_bind_key (m, key, vector, code)
     struct keymap *m;
     int key;
     int vector;
     int code;
{
  m->keys[key].vector = (short)vector;
  m->keys[key].code = (short)code;
}

static void 
interact_macro_cmd ()
{
  io_error_msg ("Command not implemented");
  return;
}

static void 
kill_cell_cmd ()
{
  CELL *cp;

  cp = find_cell (curow, cucol);
  if (!cp)
    return;
  if ((GET_LCK (cp) == LCK_DEF && default_lock == LCK_LCK) || GET_LCK (cp) == LCK_LCK)
    {
      io_error_msg ("Cell %s is locked", cell_name (curow, cucol));
      return;
    }
  new_value (curow, cucol, "");
  cp->cell_flags = 0;
  modified = 1;
}

static void
format_cell_cmd ()
{
  struct rng to;

  to.lr = to.hr = curow;
  to.lc = to.hc = cucol;
  format_area (&to);
}

static void
kill_all_cmd ()
{
  clear_spreadsheet ();
  io_repaint ();
}

static void
do_input_cmd (n, c)
     int n;
     int c;
{
  CELL *cp;
  char *fail;

  switch (n)
    {
      /* Edit cell */
    case 0:
      if (cp = find_cell (curow, cucol))
	{
	  set_line (&val_line, decomp (curow, cucol, cp));
	  decomp_free ();
	}
      else
	set_line (&val_line, "");
      break;

      /* Edit Cell Value */
    case 1:
      set_line (&val_line, cell_value_string (curow, cucol));
      break;

    case 2:			/* Set-cell */
      set_line (&val_line, "x");
      val_line.buf[0] = c;
      break;

      /* New Def Cell */
    case 3:
      break;
#ifdef TEST
    default:
      panic ("Unknown value in do_input_cell (%d)", n);
#endif
    }
  cp = find_cell (curow, cucol);
  if (((!cp || GET_LCK (cp) == LCK_DEF) && default_lock == LCK_LCK) || (cp && GET_LCK (cp) == LCK_LCK))
    {
      io_error_msg ("Cell %s is locked", cell_name (curow, cucol));
      return;
    }
  setrow = curow;
  setcol = cucol;
  if (io_get_line ("set-cell", &val_line) > 1)
    return;
  fail = new_value (setrow, setcol, val_line.buf);
  if (fail)
    io_error_msg (fail);
  else
    modified = 1;
}

static void
do_break_cmd ()
{
  if (mkrow != NON_ROW)
    {
      mkrow = NON_ROW;
      how_many = 1;
      io_update_status ();
    }
}

static void 
digit_cmd (magic)
     int magic;
{
  struct keymap *map;

  map = the_maps[DIGIT_MAP];
  map->map_next = the_maps[magic];
  how_many = 0;
  do
    {
      how_many = how_many * 10 + cur_cmd - digit_0_cmd;
      io_update_status ();
      map_chr (DIGIT_MAP);
    }
  while (cur_cmd >= digit_0_cmd && cur_cmd <= digit_9_cmd);
}




static void 
set_area_font_cmd (name, rng)
     char *name;
     struct rng *rng;
{
  struct font_memo *f = intern_font (name);
  set_area_font (rng, f);
}


static void 
set_cell_font_cmd (name)
     char *name;
{
  struct rng to;

  to.lr = to.hr = curow;
  to.lc = to.hc = cucol;
  set_area_font_cmd (name, &to);
}

static float default_pswid = 8.5 * 72.;
static float default_pshgt = 11. * 72.;

static void
set_page_size_cmd (whole_str)
     char * whole_str;
{
  char * str = whole_str;
  float neww;
  float newh;
  while (isspace(*str++)) ;
  if (!isdigit (*str))
    {
      io_error_msg ("Bad page size (should look like `8.5 x 11'): %s.",
		    whole_str);
      return;
    }
  neww = atof (str);
  while (isdigit(*str++)) ;
  if (*str == '.')
    {
      while (isdigit (*++str)) ;
      --str;
    }
  while (isspace(*str++)) ;
  if (!isdigit (*str))
    {
      io_error_msg ("Bad page size (should look like `8.5 x 11'): %s.",
		    whole_str);
      return;
    }
  newh = atof (str);
  while (isdigit(*str++)) ;
  if (*str == '.')
    {
      while (isdigit (*++str)) ;
      --str;
    }
  while (isspace(*str++)) ;
  if (*str == 'c')
    {
      neww *= .3937;
      newh *= .3937;
    }
  default_pswid = neww;
  default_pshgt = newh;
  
}

void 
psprint_region_cmd (fp, rng)
     FILE *fp;
     struct rng *rng;
{
  psprint_region (fp, rng, default_pswid, default_pshgt, 0);
}

/* This is a bit kludgey. Input line editting has its own event loop (grr!),
 * and all of its state is private.  These mouse commands can't entirely
 * handle it when the target is in the input line.  In that case, they
 * save the decoded mouse event where io_get_line can pick it up:
 */
struct mouse_event last_mouse_event;

static void
do_mouse_goto ()
{
  if (last_mouse_event.location >= 0 && last_mouse_event.downp)
    {
      nicely_goto_window (last_mouse_event.location);
      io_move_cell_cursor (last_mouse_event.r, last_mouse_event.c);
    }
}

static void
do_mouse_mark ()
{
  if (last_mouse_event.location >= 0 && last_mouse_event.downp)
    {
      mkrow = last_mouse_event.r;
      mkcol = last_mouse_event.c;
    }
}


static void
do_mouse_mark_and_goto ()
{
  if (last_mouse_event.location >= 0 && last_mouse_event.downp)
    {
      mkrow = last_mouse_event.r;
      mkcol = last_mouse_event.c;
      nicely_goto_window (last_mouse_event.location);
      io_move_cell_cursor (last_mouse_event.r, last_mouse_event.c);
    }
}

static void
do_mouse_cmd (fn)
     void (*fn) ();
{
  int seq = real_get_chr ();
  dequeue_mouse_event (&last_mouse_event, seq);
  fn ();
}

static void
mouse_mark_cmd ()
{
  do_mouse_cmd (do_mouse_mark);
}


int was_mouse_goto = 0;
static void
mouse_goto_cmd ()
{
  do_mouse_cmd (do_mouse_goto);
  was_mouse_goto = 1;
}

static void
mouse_mark_and_goto_cmd ()
{
  do_mouse_cmd (do_mouse_mark_and_goto);
}
